# -*- coding: utf-8 -*-
"""Core TrxFile class with memory-mapped data access."""

from copy import deepcopy
import json
import logging
import os
import shutil
import struct
from typing import Any, List, Optional, Tuple, Type, Union
import zipfile

import nibabel as nib
from nibabel.affines import voxel_sizes
from nibabel.nifti1 import Nifti1Header, Nifti1Image
from nibabel.orientations import aff2axcodes
from nibabel.streamlines.array_sequence import ArraySequence
from nibabel.streamlines.tractogram import LazyTractogram, Tractogram
from nibabel.streamlines.trk import TrkFile
import numpy as np

from trx.io import get_trx_tmp_dir
from trx.utils import (
    append_generator_to_dict,
    close_or_delete_mmap,
    convert_data_dict_to_tractogram,
    get_reference_info_wrapper,
)

try:
    import dipy  # noqa: F401

    dipy_available = True
except ImportError:
    dipy_available = False


def _get_dtype_little_endian(dtype: Union[np.dtype, str, type]) -> np.dtype:
    """Convert a dtype to its little-endian equivalent.

    The TRX file format uses little-endian byte order for cross-platform
    compatibility. This function ensures that dtypes are always interpreted
    as little-endian when reading/writing TRX files.

    Parameters
    ----------
    dtype : np.dtype, str, or type
        Input dtype specification (e.g., np.float32, 'float32', '>f4').

    Returns
    -------
    np.dtype
        Little-endian dtype. For single-byte types (uint8, int8, bool),
        returns the original dtype as endianness is not applicable.
    """
    dt = np.dtype(dtype)
    # Single-byte types don't have endianness
    if dt.byteorder == "|" or dt.itemsize == 1:
        return dt
    # Already little-endian
    if dt.byteorder == "<":
        return dt
    # Convert to little-endian
    return dt.newbyteorder("<")


def _ensure_little_endian(arr: np.ndarray) -> np.ndarray:
    """Ensure array data is in little-endian byte order for writing.

    Parameters
    ----------
    arr : np.ndarray
        Input array.

    Returns
    -------
    np.ndarray
        Array with little-endian byte order. Returns a copy if conversion
        was needed, otherwise returns the original array.
    """
    dt = arr.dtype
    # Single-byte types don't have endianness
    if dt.byteorder == "|" or dt.itemsize == 1:
        return arr
    # Already little-endian
    if dt.byteorder == "<":
        return arr
    # Native byte order on little-endian system
    if dt.byteorder == "=" and np.little_endian:
        return arr
    # Convert to little-endian
    return arr.astype(dt.newbyteorder("<"))


def _append_last_offsets(nib_offsets: np.ndarray, nb_vertices: int) -> np.ndarray:
    """Append the last element of offsets from header information.

    Parameters
    ----------
    nib_offsets : np.ndarray
        Array of offsets with the last element being the start of the last
        streamline (nibabel convention).
    nb_vertices : int
        Total number of vertices in the streamlines.

    Returns
    -------
    np.ndarray
        Offsets array (VTK convention).
    """

    def is_sorted(a):
        """Return True if array is sorted non-decreasing.

        Parameters
        ----------
        a : np.ndarray
            1D array of numeric offsets.

        Returns
        -------
        bool
            True when ``a`` is monotonically non-decreasing.
        """
        return np.all(a[:-1] <= a[1:])

    if not is_sorted(nib_offsets):
        raise ValueError("Offsets must be sorted values.")
    return np.append(nib_offsets, nb_vertices).astype(nib_offsets.dtype)


def _generate_filename_from_data(arr: np.ndarray, filename: str) -> str:
    """Determine the data type from array data and generate the appropriate filename.

    Parameters
    ----------
    arr : np.ndarray
        A NumPy array (1-2D, otherwise ValueError raised).
    filename : str
        The original filename.

    Returns
    -------
    str
        An updated filename with appropriate extension.
    """
    base, ext = os.path.splitext(filename)
    if ext:
        logging.warning("Will overwrite provided extension if needed.")

    dtype = arr.dtype
    dtype = "bit" if dtype is np.dtype(bool) else dtype.name

    if arr.ndim == 1:
        new_filename = "{}.{}".format(base, dtype)
    elif arr.ndim == 2:
        dim = arr.shape[-1]
        if dim == 1:
            new_filename = "{}.{}".format(base, dtype)
        else:
            new_filename = "{}.{}.{}".format(base, arr.shape[-1], dtype)
    else:
        raise ValueError("Invalid dimensionality.")

    return new_filename


def _split_ext_with_dimensionality(filename: str) -> Tuple[str, int, str]:
    """Take a filename and split it into its components.

    Parameters
    ----------
    filename : str
        Input filename.

    Returns
    -------
    tuple
        A tuple of (basename, dimension, extension).
    """
    basename = os.path.basename(filename)
    split = basename.split(".")

    if len(split) != 2 and len(split) != 3:
        raise ValueError("Invalid filename.")
    basename = split[0]
    ext = ".{}".format(split[-1])
    dim = 1 if len(split) == 2 else split[1]

    _is_dtype_valid(ext)

    return basename, int(dim), ext


def _compute_lengths(offsets: np.ndarray) -> np.ndarray:
    """Compute lengths from offsets.

    Parameters
    ----------
    offsets : np.ndarray
        An array of offsets.

    Returns
    -------
    np.ndarray
        An array of lengths.
    """
    if len(offsets) > 0:
        last_elem_pos = _dichotomic_search(offsets)
        lengths = np.ediff1d(offsets)
        if len(lengths) > last_elem_pos:
            lengths[last_elem_pos] = 0
    else:
        lengths = np.array([0])

    return lengths.astype(np.uint32)


def _is_dtype_valid(ext: str) -> bool:
    """Verify that filename extension is a valid datatype.

    Parameters
    ----------
    ext : str
        Filename extension.

    Returns
    -------
    bool
        True if the provided datatype is valid, False otherwise.
    """
    if ext.replace(".", "") == "bit":
        return True
    try:
        isinstance(np.dtype(ext.replace(".", "")), np.dtype)
        return True
    except TypeError:
        return False


def _dichotomic_search(
    x: np.ndarray, l_bound: Optional[int] = None, r_bound: Optional[int] = None
) -> int:
    """Find where data of a contiguous array is actually ending.

    Parameters
    ----------
    x : np.ndarray
        Array of values.
    l_bound : int, optional
        Lower bound index for search.
    r_bound : int, optional
        Upper bound index for search.

    Returns
    -------
    int
        Index at which array value is 0 (if possible), otherwise returns -1.
    """
    if l_bound is None and r_bound is None:
        l_bound = 0
        r_bound = len(x) - 1

    if l_bound == r_bound:
        val = l_bound if x[l_bound] != 0 else -1
        return val

    mid_bound = (l_bound + r_bound + 1) // 2

    if x[mid_bound] == 0:
        return _dichotomic_search(x, l_bound, mid_bound - 1)
    else:
        return _dichotomic_search(x, mid_bound, r_bound)


def _create_memmap(
    filename: str,
    mode: str = "r",
    shape: Tuple = (1,),
    dtype: np.dtype = np.float32,
    offset: int = 0,
    order: str = "C",
) -> np.ndarray:
    """Wrap memmap creation to support empty arrays.

    Parameters
    ----------
    filename : str
        Filename where the empty memmap should be created.
    mode : str, optional
        File open mode (see np.memmap for options). Default is 'r'.
    shape : tuple, optional
        Shape of memmapped array. Default is (1,).
    dtype : np.dtype, optional
        Datatype of memmapped array. Default is np.float32.
    offset : int, optional
        Offset of the data within the file. Default is 0.
    order : str, optional
        Data representation on disk ('C' or 'F'). Default is 'C'.

    Returns
    -------
    np.ndarray
        Memory-mapped array or a zero-filled array if shape[0] is 0.
    """
    if np.dtype(dtype) == bool:
        filename = filename.replace(".bool", ".bit")

    # TRX format uses little-endian byte order for cross-platform compatibility
    dtype = _get_dtype_little_endian(dtype)

    if shape[0]:
        return np.memmap(
            filename, mode=mode, offset=offset, shape=shape, dtype=dtype, order=order
        )
    else:
        if not os.path.isfile(filename):
            f = open(filename, "wb")
            f.close()
        return np.zeros(shape, dtype=dtype)


def load(input_obj: str, check_dpg: bool = True) -> Type["TrxFile"]:
    """Load a TrxFile (compressed or not).

    Parameters
    ----------
    input_obj : str
        A directory name or filepath to the TRX data.
    check_dpg : bool, optional
        Whether to check group metadata. Default is True.

    Returns
    -------
    TrxFile
        TrxFile object representing the read data.
    """
    # TODO Check if 0 streamlines, then 0 vertices is expected (vice-versa)
    # TODO 4x4 affine matrices should contains values (no all-zeros)
    # TODO 3x1 dimensions array should contains values at each position (int)
    if os.path.isfile(input_obj):
        was_compressed = False
        with zipfile.ZipFile(input_obj, "r") as zf:
            for info in zf.infolist():
                if info.compress_type != 0:
                    was_compressed = True
                    break
        if was_compressed:
            with zipfile.ZipFile(input_obj, "r") as zf:
                tmp_dir = get_trx_tmp_dir()
                zf.extractall(tmp_dir.name)
                trx = load_from_directory(tmp_dir.name)
                trx._uncompressed_folder_handle = tmp_dir
                logging.info(
                    "File was compressed, call the close() function before exiting."
                )
        else:
            trx = load_from_zip(input_obj)
    elif os.path.isdir(input_obj):
        trx = load_from_directory(input_obj)
    else:
        raise ValueError("File/Folder does not exist")

    # Example of robust check for metadata
    if check_dpg:
        for dpg in trx.data_per_group.keys():
            if dpg not in trx.groups.keys():
                raise ValueError(
                    "An undeclared group ({}) has data_per_group.".format(dpg)
                )
    return trx


def load_from_zip(filename: str) -> Type["TrxFile"]:
    """Load a TrxFile from a single zipfile.

    Note: Does not work with compressed zipfiles.

    Parameters
    ----------
    filename : str
        Path of the zipped TrxFile.

    Returns
    -------
    TrxFile
        TrxFile representing the read data.
    """
    with zipfile.ZipFile(filename, mode="r") as zf:
        with zf.open("header.json") as zf_header:
            header = json.load(zf_header)
            header["VOXEL_TO_RASMM"] = np.reshape(
                header["VOXEL_TO_RASMM"], (4, 4)
            ).astype(np.float32)
            header["DIMENSIONS"] = np.array(header["DIMENSIONS"], dtype=np.uint16)

        files_pointer_size = {}
        for zip_info in zf.filelist:
            elem_filename = zip_info.filename
            _, ext = os.path.splitext(elem_filename)
            if ext == ".json" or zip_info.is_dir():
                continue

            if not _is_dtype_valid(ext):
                continue
                raise ValueError("The dtype {} is not supported".format(elem_filename))

            if ext == ".bit":
                ext = ".bool"

            # Read actual local file header to get correct data offset.
            # We can't use zip_info.FileHeader() because ZIP spec allows local
            # headers to differ from central directory entries.
            # See: https://pkware.cachefly.net/webdocs/casestudies/APPNOTE.TXT
            _ZIP_LOCAL_HEADER_SIZE = 30
            _ZIP_LOCAL_HEADER_SIGNATURE = b"PK\x03\x04"

            zf.fp.seek(zip_info.header_offset)
            local_header = zf.fp.read(_ZIP_LOCAL_HEADER_SIZE)
            if len(local_header) < _ZIP_LOCAL_HEADER_SIZE:
                raise ValueError(f"Truncated local file header for {elem_filename}")
            if local_header[:4] != _ZIP_LOCAL_HEADER_SIGNATURE:
                raise ValueError(
                    f"Invalid local file header signature for {elem_filename}"
                )
            fname_len, extra_len = struct.unpack("<HH", local_header[26:30])

            mem_adress = (
                zip_info.header_offset + _ZIP_LOCAL_HEADER_SIZE + fname_len + extra_len
            )

            dtype_size = np.dtype(ext[1:]).itemsize
            size = zip_info.file_size / dtype_size

            if size.is_integer():
                files_pointer_size[elem_filename] = mem_adress, int(size)
            else:
                raise ValueError("Wrong size or datatype")

    return TrxFile._create_trx_from_pointer(
        header, files_pointer_size, root_zip=filename
    )


def load_from_directory(directory: str) -> Type["TrxFile"]:
    """Load a TrxFile from a folder containing memmaps.

    Parameters
    ----------
    directory : str
        Path of the directory containing TRX data.

    Returns
    -------
    TrxFile
        TrxFile representing the read data.
    """

    directory = os.path.abspath(directory)
    with open(os.path.join(directory, "header.json")) as header:
        header = json.load(header)
        header["VOXEL_TO_RASMM"] = np.reshape(header["VOXEL_TO_RASMM"], (4, 4)).astype(
            np.float32
        )
        header["DIMENSIONS"] = np.array(header["DIMENSIONS"], dtype=np.uint16)
    files_pointer_size = {}
    for root, _dirs, files in os.walk(directory):
        for name in files:
            elem_filename = os.path.join(root, name)
            _, ext = os.path.splitext(elem_filename)
            if ext == ".json":
                continue

            if not _is_dtype_valid(ext):
                raise ValueError(
                    "The dtype of {} is not supported".format(elem_filename)
                )

            if ext == ".bit":
                ext = ".bool"

            dtype_size = np.dtype(ext[1:]).itemsize
            size = os.path.getsize(elem_filename) / dtype_size
            if size.is_integer():
                files_pointer_size[elem_filename] = 0, int(size)
            elif os.path.getsize(elem_filename) == 1:
                files_pointer_size[elem_filename] = 0, 0
            else:
                raise ValueError("Wrong size or datatype")

    return TrxFile._create_trx_from_pointer(header, files_pointer_size, root=directory)


def _filter_empty_trx_files(trx_list: List["TrxFile"]) -> List["TrxFile"]:
    """Remove empty TrxFiles from the list.

    Parameters
    ----------
    trx_list : list of TrxFile class instances
        Collection of tractograms to filter.

    Returns
    -------
    list of TrxFile class instances
        Only entries containing at least one streamline.
    """
    return [curr_trx for curr_trx in trx_list if curr_trx.header["NB_STREAMLINES"] > 0]


def _get_all_data_keys(trx_list: List["TrxFile"]) -> Tuple[set, set]:
    """Get all dps and dpv keys from the TrxFile list.

    Parameters
    ----------
    trx_list : list of TrxFile class instances
        Collection of tractograms.

    Returns
    -------
    tuple of set
        Sets of `data_per_streamline` keys and `data_per_vertex` keys.
    """
    all_dps = []
    all_dpv = []
    for curr_trx in trx_list:
        all_dps.extend(list(curr_trx.data_per_streamline.keys()))
        all_dpv.extend(list(curr_trx.data_per_vertex.keys()))
    return set(all_dps), set(all_dpv)


def _check_space_attributes(trx_list: List["TrxFile"]) -> None:
    """Verify that space attributes are consistent across TrxFiles.

    Parameters
    ----------
    trx_list : list of TrxFile
        Tractograms to compare for affine and dimension consistency.

    Raises
    ------
    ValueError
        If voxel-to-RASMM matrices or dimensions differ.
    """
    ref_trx = trx_list[0]
    for curr_trx in trx_list[1:]:
        if not np.allclose(
            ref_trx.header["VOXEL_TO_RASMM"], curr_trx.header["VOXEL_TO_RASMM"]
        ) or not np.array_equal(
            ref_trx.header["DIMENSIONS"], curr_trx.header["DIMENSIONS"]
        ):
            raise ValueError("Wrong space attributes.")


def _verify_dpv_coherence(
    trx_list: List["TrxFile"], all_dpv: set, ref_trx: "TrxFile", delete_dpv: bool
) -> None:
    """Verify dpv coherence across TrxFiles.

    Parameters
    ----------
    trx_list : list of TrxFile class instances
        Tractograms being concatenated.
    all_dpv : set
        Union of `data_per_vertex` keys across tractograms.
    ref_trx : TrxFile class instance
        Reference tractogram for dtype/key checks.
    delete_dpv : bool
        Drop mismatched dpv keys instead of raising when True.

    Raises
    ------
    ValueError
        If dpv keys or dtypes differ and `delete_dpv` is False.
    """
    for curr_trx in trx_list:
        for key in all_dpv:
            if (
                key not in ref_trx.data_per_vertex.keys()
                or key not in curr_trx.data_per_vertex.keys()
            ):
                if not delete_dpv:
                    logging.debug(
                        "{} dpv key does not exist in all TrxFile.".format(key)
                    )
                    raise ValueError("TrxFile must be sharing identical dpv keys.")
            elif (
                ref_trx.data_per_vertex[key]._data.dtype
                != curr_trx.data_per_vertex[key]._data.dtype
            ):
                logging.debug(
                    "{} dpv key is not declared with the same dtype "
                    "in all TrxFile.".format(key)
                )
                raise ValueError("Shared dpv key, has different dtype.")


def _verify_dps_coherence(
    trx_list: List["TrxFile"], all_dps: set, ref_trx: "TrxFile", delete_dps: bool
) -> None:
    """Verify dps coherence across TrxFiles.

    Parameters
    ----------
    trx_list : list of TrxFile class instances
        Tractograms being concatenated.
    all_dps : set
        Union of data_per_streamline keys across tractograms.
    ref_trx : TrxFile class instance
        Reference tractogram for dtype/key checks.
    delete_dps : bool
        Drop mismatched dps keys instead of raising when True.

    Raises
    ------
    ValueError
        If dps keys or dtypes differ and `delete_dps` is False.
    """
    for curr_trx in trx_list:
        for key in all_dps:
            if (
                key not in ref_trx.data_per_streamline.keys()
                or key not in curr_trx.data_per_streamline.keys()
            ):
                if not delete_dps:
                    logging.debug(
                        "{} dps key does not exist in all TrxFile.".format(key)
                    )
                    raise ValueError("TrxFile must be sharing identical dps keys.")
            elif (
                ref_trx.data_per_streamline[key].dtype
                != curr_trx.data_per_streamline[key].dtype
            ):
                logging.debug(
                    "{} dps key is not declared with the same dtype "
                    "in all TrxFile.".format(key)
                )
                raise ValueError("Shared dps key, has different dtype.")


def _compute_groups_info(trx_list: List["TrxFile"]) -> Tuple[dict, dict]:
    """Compute group length and dtype information.

    Parameters
    ----------
    trx_list : list of TrxFile class instances
        Tractograms being concatenated.

    Returns
    -------
    tuple of dict
        (group lengths, group dtypes) keyed by group name.
    """
    all_groups_len = {}
    all_groups_dtype = {}

    for trx_1 in trx_list:
        for group_key in trx_1.groups.keys():
            if group_key in all_groups_len:
                all_groups_len[group_key] += len(trx_1.groups[group_key])
            else:
                all_groups_len[group_key] = len(trx_1.groups[group_key])

            if (
                group_key in all_groups_dtype
                and trx_1.groups[group_key].dtype != all_groups_dtype[group_key]
            ):
                raise ValueError("Shared group key, has different dtype.")
            else:
                all_groups_dtype[group_key] = trx_1.groups[group_key].dtype

    return all_groups_len, all_groups_dtype


def _create_new_trx_for_concatenation(
    trx_list: List["TrxFile"],
    ref_trx: "TrxFile",
    delete_dps: bool,
    delete_dpv: bool,
    delete_groups: bool,
) -> "TrxFile":
    """Create a new TrxFile for concatenation.

    Parameters
    ----------
    trx_list : list of TrxFile class instances
        Input tractograms to concatenate.
    ref_trx : TrxFile class instance
        Reference tractogram for header/dtype template.
    delete_dps : bool
        Drop `data_per_streamline` keys not shared.
    delete_dpv : bool
        Drop `data_per_vertex` keys not shared.
    delete_groups : bool
        Drop groups when metadata differ.

    Returns
    -------
    TrxFile
        Empty TRX ready to receive concatenated data.
    """
    nb_vertices = 0
    nb_streamlines = 0
    for curr_trx in trx_list:
        curr_strs_len, curr_pts_len = curr_trx._get_real_len()
        nb_streamlines += curr_strs_len
        nb_vertices += curr_pts_len

    new_trx = TrxFile(
        nb_vertices=nb_vertices, nb_streamlines=nb_streamlines, init_as=ref_trx
    )
    if delete_dps:
        new_trx.data_per_streamline = {}
    if delete_dpv:
        new_trx.data_per_vertex = {}
    if delete_groups:
        new_trx.groups = {}

    return new_trx


def _setup_groups_for_concatenation(
    new_trx: "TrxFile",
    trx_list: List["TrxFile"],
    all_groups_len: dict,
    all_groups_dtype: dict,
    delete_groups: bool,
) -> None:
    """Setup groups in the new TrxFile for concatenation.

    Parameters
    ----------
    new_trx : TrxFile class instance
        Destination tractogram.
    trx_list : list of TrxFile class instances
        Source tractograms.
    all_groups_len : dict
        Mapping of group name to total length.
    all_groups_dtype : dict
        Mapping of group name to dtype.
    delete_groups : bool
        If True, skip creating group arrays.
    """
    if delete_groups:
        return

    tmp_dir = new_trx._uncompressed_folder_handle.name

    for group_key in all_groups_len.keys():
        if not os.path.isdir(os.path.join(tmp_dir, "groups/")):
            os.mkdir(os.path.join(tmp_dir, "groups/"))

        dtype = all_groups_dtype[group_key]
        group_filename = os.path.join(
            tmp_dir, "groups/{}.{}".format(group_key, dtype.name)
        )
        group_len = all_groups_len[group_key]
        new_trx.groups[group_key] = _create_memmap(
            group_filename, mode="w+", shape=(group_len,), dtype=dtype
        )

        pos = 0
        count = 0
        for curr_trx in trx_list:
            curr_len = len(curr_trx.groups[group_key])
            new_trx.groups[group_key][pos : pos + curr_len] = (
                curr_trx.groups[group_key] + count
            )
            pos += curr_len
            count += curr_trx.header["NB_STREAMLINES"]


def concatenate(
    trx_list: List["TrxFile"],
    delete_dpv: bool = False,
    delete_dps: bool = False,
    delete_groups: bool = False,
    check_space_attributes: bool = True,
    preallocation: bool = False,
) -> "TrxFile":
    """Concatenate multiple TrxFile together, with support for preallocation.

    Parameters
    ----------
    trx_list : list of TrxFile
        A list containing TrxFiles to concatenate.
    delete_dpv : bool, optional
        Delete dpv keys that do not exist in all the provided TrxFiles.
        Default is False.
    delete_dps : bool, optional
        Delete dps keys that do not exist in all the provided TrxFiles.
        Default is False.
    delete_groups : bool, optional
        Delete all the groups that currently exist in the TrxFiles.
        Default is False.
    check_space_attributes : bool, optional
        Verify that dimensions and size of data are similar between all
        the TrxFiles. Default is True.
    preallocation : bool, optional
        Preallocated TrxFile has already been generated and is the first
        element in trx_list. Note: delete_groups must be set to True as well.
        Default is False.

    Returns
    -------
    TrxFile
        TrxFile representing the concatenated data.
    """
    trx_list = _filter_empty_trx_files(trx_list)
    if len(trx_list) == 0:
        logging.warning("Inputs of concatenation were empty.")
        return TrxFile()

    ref_trx = trx_list[0]
    all_dps, all_dpv = _get_all_data_keys(trx_list)

    if check_space_attributes:
        _check_space_attributes(trx_list)

    if preallocation and not delete_groups:
        raise ValueError("Groups are variables, cannot be handled with preallocation")

    _verify_dpv_coherence(trx_list, all_dpv, ref_trx, delete_dpv)
    _verify_dps_coherence(trx_list, all_dps, ref_trx, delete_dps)

    all_groups_len, all_groups_dtype = _compute_groups_info(trx_list)

    to_concat_list = trx_list[1:] if preallocation else trx_list
    if not preallocation:
        new_trx = _create_new_trx_for_concatenation(
            to_concat_list, ref_trx, delete_dps, delete_dpv, delete_groups
        )
        _setup_groups_for_concatenation(
            new_trx, trx_list, all_groups_len, all_groups_dtype, delete_groups
        )
        strs_end, pts_end = 0, 0
    else:
        new_trx = ref_trx
        strs_end, pts_end = new_trx._get_real_len()

    for curr_trx in to_concat_list:
        strs_end, pts_end = new_trx._copy_fixed_arrays_from(
            curr_trx, strs_start=strs_end, pts_start=pts_end
        )
    return new_trx


def save(
    trx: "TrxFile", filename: str, compression_standard: Any = zipfile.ZIP_STORED
) -> None:
    """Save a TrxFile (compressed or not).

    Parameters
    ----------
    trx : TrxFile
        The TrxFile to save.
    filename : str
        The path to save the TrxFile to.
    compression_standard : int, optional
        The compression standard to use, as defined by the ZipFile library.
        Default is zipfile.ZIP_STORED.
    """
    _, ext = os.path.splitext(filename)
    if ext not in [".zip", ".trx", ""]:
        raise ValueError("Unsupported extension.")

    copy_trx = trx.deepcopy()
    copy_trx.resize()
    tmp_dir_name = copy_trx._uncompressed_folder_handle.name
    if ext in [".zip", ".trx"]:
        zip_from_folder(tmp_dir_name, filename, compression_standard)
    else:
        if os.path.isdir(filename):
            shutil.rmtree(filename)
        shutil.copytree(tmp_dir_name, filename)
    copy_trx.close()


def zip_from_folder(
    directory: str, filename: str, compression_standard: Any = zipfile.ZIP_STORED
) -> None:
    """Zip on-disk memmaps into a single file.

    Parameters
    ----------
    directory : str
        The path to the on-disk memmap directory.
    filename : str
        The path where the zip file should be created.
    compression_standard : int, optional
        The compression standard to use, as defined by the ZipFile library.
        Default is zipfile.ZIP_STORED.
    """
    with zipfile.ZipFile(filename, mode="w", compression=compression_standard) as zf:
        for root, _, files in os.walk(directory):
            for name in files:
                curr_filename = os.path.join(root, name)
                tmp_filename = curr_filename.replace(directory, "")[1:]
                zf.write(curr_filename, tmp_filename)


class TrxFile:
    """Core class of the TrxFile.

    Parameters
    ----------
    nb_vertices : int, optional
        The number of vertices to use in the new TrxFile.
    nb_streamlines : int, optional
        The number of streamlines in the new TrxFile.
    init_as : TrxFile class instance, optional
        A TrxFile to use as reference.

    reference : str, dict, Nifti1Image, TrkFile, or Nifti1Header, optional
        A Nifti or Trk file/obj to use as reference.
    """

    header: dict
    streamlines: Type[ArraySequence]
    groups: dict
    data_per_streamline: dict
    data_per_vertex: dict
    data_per_group: dict

    def __init__(
        self,
        nb_vertices: Optional[int] = None,
        nb_streamlines: Optional[int] = None,
        init_as: Optional[Type["TrxFile"]] = None,
        reference: Union[
            str,
            dict,
            Type[Nifti1Image],
            Type[TrkFile],
            Type[Nifti1Header],
            None,
        ] = None,
    ) -> None:
        """Initialize an empty TrxFile with support for preallocation.

        Parameters
        ----------
        nb_vertices : int, optional
            The number of vertices to use in the new TrxFile.
        nb_streamlines : int, optional
            The number of streamlines in the new TrxFile.
        init_as : TrxFile, optional
            A TrxFile to use as reference.
        reference : str, dict, Nifti1Image, TrkFile, Nifti1Header, optional
            A Nifti or Trk file/obj to use as reference.
        """
        if init_as is not None:
            affine = init_as.header["VOXEL_TO_RASMM"]
            dimensions = init_as.header["DIMENSIONS"]
        elif reference is not None:
            affine, dimensions, _, _ = get_reference_info_wrapper(reference)
        else:
            logging.debug(
                "No reference provided, using blank space "
                "attributes, please update them later."
            )
            affine = np.eye(4).astype(np.float32)
            dimensions = np.array([1, 1, 1], dtype=np.uint16)

        if nb_vertices is None and nb_streamlines is None:
            if init_as is not None:
                raise ValueError(
                    "Can't use init_as without declaring nb_vertices AND nb_streamlines"
                )
            logging.debug("Initializing empty TrxFile.")
            self.header = {}
            # Using the new format default type
            tmp_strs = ArraySequence()
            tmp_strs._data = tmp_strs._data.astype(np.float32)
            tmp_strs._offsets = tmp_strs._offsets.astype(np.uint32)
            tmp_strs._lengths = tmp_strs._lengths.astype(np.uint32)
            self.streamlines = tmp_strs
            self.groups = {}
            self.data_per_streamline = {}
            self.data_per_vertex = {}
            self.data_per_group = {}
            self._uncompressed_folder_handle = None

            nb_vertices = 0
            nb_streamlines = 0

        elif nb_vertices is not None and nb_streamlines is not None:
            logging.debug(
                "Preallocating TrxFile with size {} streamlinesand {} vertices.".format(
                    nb_streamlines, nb_vertices
                )
            )
            trx = self._initialize_empty_trx(
                nb_streamlines, nb_vertices, init_as=init_as
            )
            self.__dict__ = trx.__dict__
        else:
            raise ValueError("You must declare both nb_vertices AND NB_STREAMLINES")

        self.header["VOXEL_TO_RASMM"] = affine
        self.header["DIMENSIONS"] = dimensions
        self.header["NB_VERTICES"] = nb_vertices
        self.header["NB_STREAMLINES"] = nb_streamlines
        self._copy_safe = True

    def __str__(self) -> str:
        """Generate the string for printing"""
        affine = np.array(self.header["VOXEL_TO_RASMM"], dtype=np.float32)
        dimensions = np.array(self.header["DIMENSIONS"], dtype=np.uint16)
        vox_sizes = np.array(voxel_sizes(affine), dtype=np.float32)
        vox_order = "".join(aff2axcodes(affine))

        text = "VOXEL_TO_RASMM: \n{}".format(
            np.array2string(affine, formatter={"float_kind": lambda x: "%.6f" % x})
        )
        text += "\nDIMENSIONS: {}".format(np.array2string(dimensions))
        text += "\nVOX_SIZES: {}".format(
            np.array2string(vox_sizes, formatter={"float_kind": lambda x: "%.2f" % x})
        )
        text += "\nVOX_ORDER: {}".format(vox_order)

        strs_size = self.header["NB_STREAMLINES"]
        pts_size = self.header["NB_VERTICES"]
        strs_len, pts_len = self._get_real_len()

        if strs_size != strs_len or pts_size != pts_len:
            text += "\nstreamline_size: {}".format(strs_size)
            text += "\nvertex_size: {}".format(pts_size)

        text += "\nstreamline_count: {}".format(strs_len)
        text += "\nvertex_count: {}".format(pts_len)

        dpv_keys = list(self.data_per_vertex.keys())
        if dpv_keys:
            text += "\ndata_per_vertex keys: {}".format(dpv_keys)
        else:
            text += "\nNo data per vertex (dpv) keys"

        dps_keys = list(self.data_per_streamline.keys())
        if dps_keys:
            text += "\ndata_per_streamline keys: {}".format(dps_keys)
        else:
            text += "\nNo data per streamline (dps) keys"

        group_keys = list(self.groups.keys())
        if group_keys:
            text += "\ngroups keys: {}".format(group_keys)
        else:
            text += "\nNo group keys"
        for group_key in self.groups.keys():
            if group_key in self.data_per_group:
                text += "\ndata_per_groups ({}) keys: {}".format(
                    group_key, list(self.data_per_group[group_key].keys())
                )

        text += "\ncopy_safe: {}".format(self._copy_safe)

        return text

    def __len__(self) -> int:
        """Define the length of the object"""
        return len(self.streamlines)

    def __getitem__(self, key) -> Any:
        """Slice all data in a consistent way"""
        if isinstance(key, int):
            if key < 0:
                key += len(self)
            key = [key]
        elif isinstance(key, slice):
            key = list(range(*key.indices(len(self))))

        return self.select(key, keep_group=False)

    def __deepcopy__(self) -> Type["TrxFile"]:
        """Return a deep copy of the TrxFile.

        Parameters
        ----------
        self
            TrxFile class instance.

        Returns
        -------
        TrxFile class instance
            Deep-copied instance.
        """
        return self.deepcopy()

    def deepcopy(self) -> Type["TrxFile"]:  # noqa: C901
        """Create a deepcopy of the TrxFile.

        Returns
        -------
        TrxFile
            A deepcopied TrxFile of the current TrxFile.
        """
        tmp_dir = get_trx_tmp_dir()
        out_json = open(os.path.join(tmp_dir.name, "header.json"), "w")
        tmp_header = deepcopy(self.header)

        if not isinstance(tmp_header["VOXEL_TO_RASMM"], list):
            tmp_header["VOXEL_TO_RASMM"] = tmp_header["VOXEL_TO_RASMM"].tolist()
        if not isinstance(tmp_header["DIMENSIONS"], list):
            tmp_header["DIMENSIONS"] = tmp_header["DIMENSIONS"].tolist()

        # tofile() always write in C-order
        # Ensure little-endian byte order for cross-platform compatibility
        if not self._copy_safe:
            to_dump = self.streamlines.copy()._data
            tmp_header["NB_STREAMLINES"] = len(self.streamlines)
            tmp_header["NB_VERTICES"] = len(to_dump)
        else:
            to_dump = self.streamlines._data
        json.dump(tmp_header, out_json)
        out_json.close()

        # Only write positions and offsets if TRX is not empty
        if tmp_header["NB_STREAMLINES"] > 0 and tmp_header["NB_VERTICES"] > 0:
            positions_filename = _generate_filename_from_data(
                to_dump, os.path.join(tmp_dir.name, "positions")
            )
            _ensure_little_endian(to_dump).tofile(positions_filename)

            if not self._copy_safe:
                to_dump = _append_last_offsets(
                    self.streamlines.copy()._offsets, self.header["NB_VERTICES"]
                )
            else:
                to_dump = _append_last_offsets(
                    self.streamlines._offsets, self.header["NB_VERTICES"]
                )
            offsets_filename = _generate_filename_from_data(
                self.streamlines._offsets, os.path.join(tmp_dir.name, "offsets")
            )
            _ensure_little_endian(to_dump).tofile(offsets_filename)

        if len(self.data_per_vertex.keys()) > 0:
            os.mkdir(os.path.join(tmp_dir.name, "dpv/"))
        for dpv_key in self.data_per_vertex.keys():
            if not self._copy_safe:
                to_dump = self.data_per_vertex[dpv_key].copy()._data
            else:
                to_dump = self.data_per_vertex[dpv_key]._data

            dpv_filename = _generate_filename_from_data(
                to_dump, os.path.join(tmp_dir.name, "dpv/", dpv_key)
            )
            _ensure_little_endian(to_dump).tofile(dpv_filename)

        if len(self.data_per_streamline.keys()) > 0:
            os.mkdir(os.path.join(tmp_dir.name, "dps/"))
        for dps_key in self.data_per_streamline.keys():
            to_dump = self.data_per_streamline[dps_key]
            dps_filename = _generate_filename_from_data(
                to_dump, os.path.join(tmp_dir.name, "dps/", dps_key)
            )
            _ensure_little_endian(to_dump).tofile(dps_filename)

        if len(self.groups.keys()) > 0:
            os.mkdir(os.path.join(tmp_dir.name, "groups/"))
        for group_key in self.groups.keys():
            to_dump = self.groups[group_key]
            group_filename = _generate_filename_from_data(
                to_dump, os.path.join(tmp_dir.name, "groups/", group_key)
            )
            _ensure_little_endian(to_dump).tofile(group_filename)

            if group_key not in self.data_per_group:
                continue
            for dpg_key in self.data_per_group[group_key].keys():
                # Creates 'dpg/' only if required
                if not os.path.isdir(os.path.join(tmp_dir.name, "dpg/")):
                    os.mkdir(os.path.join(tmp_dir.name, "dpg/"))
                if not os.path.isdir(os.path.join(tmp_dir.name, "dpg/", group_key)):
                    os.mkdir(os.path.join(tmp_dir.name, "dpg/", group_key))
                to_dump = self.data_per_group[group_key][dpg_key]
                dpg_filename = _generate_filename_from_data(
                    to_dump, os.path.join(tmp_dir.name, "dpg/", group_key, dpg_key)
                )
                _ensure_little_endian(to_dump).tofile(dpg_filename)

        copy_trx = load_from_directory(tmp_dir.name)
        copy_trx._uncompressed_folder_handle = tmp_dir

        return copy_trx

    def _get_real_len(self) -> Tuple[int, int]:
        """Get the real size of data (ignoring zeros of preallocation).

        Returns
        -------
        tuple of int
            A tuple (strs_end, pts_end) representing the index of the last
            streamline and the total length of all the streamlines.
        """
        if len(self.streamlines._lengths) == 0:
            return 0, 0

        last_elem_pos = _dichotomic_search(self.streamlines._lengths)
        if last_elem_pos != -1:
            strs_end = int(last_elem_pos + 1)
            pts_end = int(np.sum(self.streamlines._lengths[0:strs_end]))
            return strs_end, pts_end

        return 0, 0

    def _copy_fixed_arrays_from(
        self,
        trx: Type["TrxFile"],
        strs_start: int = 0,
        pts_start: int = 0,
        nb_strs_to_copy: Optional[int] = None,
    ) -> Tuple[int, int]:
        """Fill a TrxFile using another and start indexes (preallocation).

        Parameters
        ----------
        trx : TrxFile
            TrxFile to copy data from.
        strs_start : int, optional
            The start index of the streamline. Default is 0.
        pts_start : int, optional
            The start index of the point. Default is 0.
        nb_strs_to_copy : int, optional
            The number of streamlines to copy. If not set, will copy all.

        Returns
        -------
        tuple of int
            A tuple (strs_end, pts_end) representing the end of the copied
            streamlines and end of copied points.
        """
        if nb_strs_to_copy is None:
            curr_strs_len, curr_pts_len = trx._get_real_len()
        else:
            curr_strs_len = int(nb_strs_to_copy)
            curr_pts_len = np.sum(trx.streamlines._lengths[0:curr_strs_len])
            curr_pts_len = int(curr_pts_len)

        strs_end = strs_start + curr_strs_len
        pts_end = pts_start + curr_pts_len

        if curr_pts_len == 0:
            return strs_start, pts_start

        # Mandatory arrays
        self.streamlines._data[pts_start:pts_end] = trx.streamlines._data[
            0:curr_pts_len
        ]
        self.streamlines._offsets[strs_start:strs_end] = (
            trx.streamlines._offsets[0:curr_strs_len] + pts_start
        )
        self.streamlines._lengths[strs_start:strs_end] = trx.streamlines._lengths[
            0:curr_strs_len
        ]

        # Optional fixed-sized arrays
        for dpv_key in self.data_per_vertex.keys():
            self.data_per_vertex[dpv_key]._data[pts_start:pts_end] = (
                trx.data_per_vertex[dpv_key]._data[0:curr_pts_len]
            )
            self.data_per_vertex[dpv_key]._offsets = self.streamlines._offsets
            self.data_per_vertex[dpv_key]._lengths = self.streamlines._lengths

        for dps_key in self.data_per_streamline.keys():
            self.data_per_streamline[dps_key][strs_start:strs_end] = (
                trx.data_per_streamline[dps_key][0:curr_strs_len]
            )

        return strs_end, pts_end

    @staticmethod
    def _initialize_empty_trx(  # noqa: C901
        nb_streamlines: int,
        nb_vertices: int,
        init_as: Optional[Type["TrxFile"]] = None,
    ) -> Type["TrxFile"]:
        """Create on-disk memmaps of a certain size (preallocation).

        Parameters
        ----------
        nb_streamlines : int
            The number of streamlines that the empty TrxFile will be
            initialized with.
        nb_vertices : int
            The number of vertices that the empty TrxFile will be
            initialized with.
        init_as : TrxFile, optional
            A TrxFile to initialize the empty TrxFile with.

        Returns
        -------
        TrxFile
            An empty TrxFile preallocated with a certain size.
        """
        trx = TrxFile()
        tmp_dir = get_trx_tmp_dir()
        logging.info("Temporary folder for memmaps: {}".format(tmp_dir.name))

        trx.header["NB_VERTICES"] = nb_vertices
        trx.header["NB_STREAMLINES"] = nb_streamlines

        if init_as is not None:
            trx.header["VOXEL_TO_RASMM"] = init_as.header["VOXEL_TO_RASMM"]
            trx.header["DIMENSIONS"] = init_as.header["DIMENSIONS"]
            positions_dtype = init_as.streamlines._data.dtype
            offsets_dtype = init_as.streamlines._offsets.dtype
            lengths_dtype = init_as.streamlines._lengths.dtype
        else:
            positions_dtype = np.dtype(np.float16)
            offsets_dtype = np.dtype(np.uint32)
            lengths_dtype = np.dtype(np.uint32)

        logging.debug(
            "Initializing positions with dtype:    {}".format(positions_dtype.name)
        )
        logging.debug("Initializing offsets with dtype: {}".format(offsets_dtype.name))
        logging.debug("Initializing lengths with dtype: {}".format(lengths_dtype.name))

        # A TrxFile without init_as only contain the essential arrays
        positions_filename = os.path.join(
            tmp_dir.name, "positions.3.{}".format(positions_dtype.name)
        )
        trx.streamlines._data = _create_memmap(
            positions_filename, mode="w+", shape=(nb_vertices, 3), dtype=positions_dtype
        )

        offsets_filename = os.path.join(
            tmp_dir.name, "offsets.{}".format(offsets_dtype.name)
        )
        trx.streamlines._offsets = _create_memmap(
            offsets_filename, mode="w+", shape=(nb_streamlines,), dtype=offsets_dtype
        )
        trx.streamlines._lengths = np.zeros(
            shape=(nb_streamlines,), dtype=lengths_dtype
        )

        # Only the structure of fixed-size arrays is copied
        if init_as is not None:
            if len(init_as.data_per_vertex.keys()) > 0:
                os.mkdir(os.path.join(tmp_dir.name, "dpv/"))
            if len(init_as.data_per_streamline.keys()) > 0:
                os.mkdir(os.path.join(tmp_dir.name, "dps/"))

            for dpv_key in init_as.data_per_vertex.keys():
                dtype = init_as.data_per_vertex[dpv_key]._data.dtype
                tmp_as = init_as.data_per_vertex[dpv_key]._data
                if tmp_as.ndim == 1:
                    dpv_filename = os.path.join(
                        tmp_dir.name, "dpv/{}.{}".format(dpv_key, dtype.name)
                    )
                    shape = (nb_vertices, 1)
                elif tmp_as.ndim == 2:
                    dim = tmp_as.shape[-1]
                    shape = (nb_vertices, dim)
                    dpv_filename = os.path.join(
                        tmp_dir.name, "dpv/{}.{}.{}".format(dpv_key, dim, dtype.name)
                    )
                else:
                    raise ValueError("Invalid dimensionality.")

                logging.debug(
                    "Initializing {} (dpv) with dtype: {}".format(dpv_key, dtype.name)
                )
                trx.data_per_vertex[dpv_key] = ArraySequence()
                trx.data_per_vertex[dpv_key]._data = _create_memmap(
                    dpv_filename, mode="w+", shape=shape, dtype=dtype
                )
                trx.data_per_vertex[dpv_key]._offsets = trx.streamlines._offsets
                trx.data_per_vertex[dpv_key]._lengths = trx.streamlines._lengths

            for dps_key in init_as.data_per_streamline.keys():
                dtype = init_as.data_per_streamline[dps_key].dtype
                tmp_as = init_as.data_per_streamline[dps_key]
                if tmp_as.ndim == 1:
                    dps_filename = os.path.join(
                        tmp_dir.name, "dps/{}.{}".format(dps_key, dtype.name)
                    )
                    shape = (nb_streamlines,)
                elif tmp_as.ndim == 2:
                    dim = tmp_as.shape[-1]
                    shape = (nb_streamlines, dim)
                    dps_filename = os.path.join(
                        tmp_dir.name, "dps/{}.{}.{}".format(dps_key, dim, dtype.name)
                    )
                else:
                    raise ValueError("Invalid dimensionality.")

                logging.debug(
                    "Initializing {} (dps) with and dtype: {}".format(
                        dps_key, dtype.name
                    )
                )
                trx.data_per_streamline[dps_key] = _create_memmap(
                    dps_filename, mode="w+", shape=shape, dtype=dtype
                )

        trx._uncompressed_folder_handle = tmp_dir

        return trx

    def _create_trx_from_pointer(  # noqa: C901
        header: dict,
        dict_pointer_size: dict,
        root_zip: Optional[str] = None,
        root: Optional[str] = None,
    ) -> Type["TrxFile"]:
        """Create a TrxFile after reading the structure of a zip/folder.

        Parameters
        ----------
        header : dict
            A TrxFile header dictionary which will be used for the new TrxFile.
        dict_pointer_size : dict
            A dictionary containing the filenames of all the files within the
            TrxFile disk file/folder.
        root_zip : str, optional
            The path of the ZipFile pointer.
        root : str, optional
            The dirname of the ZipFile pointer.

        Returns
        -------
        TrxFile
            A TrxFile constructed from the pointer provided.
        """
        trx = TrxFile()
        trx.header = header

        # Handle empty TRX files early - no positions/offsets to load
        if header["NB_STREAMLINES"] == 0 or header["NB_VERTICES"] == 0:
            return trx

        positions, offsets = None, None
        for elem_filename in dict_pointer_size.keys():
            if root_zip:
                filename = root_zip
            else:
                filename = elem_filename

            folder = os.path.dirname(elem_filename)
            base, dim, ext = _split_ext_with_dimensionality(elem_filename)
            if ext == ".bit":
                ext = ".bool"
            mem_adress, size = dict_pointer_size[elem_filename]

            if root is not None:
                # This is for Unix
                if os.name != "nt" and folder.startswith(root.rstrip("/")):
                    folder = folder.replace(root, "").lstrip("/")
                # These three are for Windows
                elif os.path.isdir(folder) and os.path.basename(folder) in [
                    "dpv",
                    "dps",
                    "groups",
                ]:
                    folder = os.path.basename(folder)
                elif os.path.basename(os.path.dirname(folder)) == "dpg":
                    folder = os.path.join("dpg", os.path.basename(folder))
                else:
                    folder = ""

            # Parse/walk the directory tree
            if base == "positions" and folder == "":
                if size != trx.header["NB_VERTICES"] * 3 or dim != 3:
                    raise ValueError("Wrong data size/dimensionality.")
                positions = _create_memmap(
                    filename,
                    mode="r+",
                    offset=mem_adress,
                    shape=(trx.header["NB_VERTICES"], 3),
                    dtype=ext[1:],
                )
            elif base == "offsets" and folder == "":
                if size != trx.header["NB_STREAMLINES"] + 1 or dim != 1:
                    raise ValueError("Wrong offsets size/dimensionality.")
                offsets = _create_memmap(
                    filename,
                    mode="r+",
                    offset=mem_adress,
                    shape=(trx.header["NB_STREAMLINES"] + 1,),
                    dtype=ext[1:],
                )
                if offsets[-1] != 0:
                    lengths = _compute_lengths(offsets)
                else:
                    lengths = [0]
            elif folder == "dps":
                nb_scalar = size / trx.header["NB_STREAMLINES"]
                if not nb_scalar.is_integer() or nb_scalar != dim:
                    raise ValueError("Wrong dps size/dimensionality.")
                else:
                    shape = (trx.header["NB_STREAMLINES"], int(nb_scalar))

                trx.data_per_streamline[base] = _create_memmap(
                    filename, mode="r+", offset=mem_adress, shape=shape, dtype=ext[1:]
                )
            elif folder == "dpv":
                nb_scalar = size / trx.header["NB_VERTICES"]
                if not nb_scalar.is_integer() or nb_scalar != dim:
                    raise ValueError("Wrong dpv size/dimensionality.")
                else:
                    shape = (trx.header["NB_VERTICES"], int(nb_scalar))

                trx.data_per_vertex[base] = _create_memmap(
                    filename, mode="r+", offset=mem_adress, shape=shape, dtype=ext[1:]
                )
            elif folder.startswith("dpg"):
                if int(size) != dim:
                    raise ValueError("Wrong dpg size/dimensionality.")
                else:
                    shape = (1, int(size))

                # Handle the two-layers architecture
                data_name = os.path.basename(base)
                sub_folder = os.path.basename(folder)
                if sub_folder not in trx.data_per_group:
                    trx.data_per_group[sub_folder] = {}
                trx.data_per_group[sub_folder][data_name] = _create_memmap(
                    filename, mode="r+", offset=mem_adress, shape=shape, dtype=ext[1:]
                )
            elif folder == "groups":
                # Groups are simply indices, nothing else
                # TODO Crash if not uint?
                if dim != 1:
                    raise ValueError("Wrong group dimensionality.")
                else:
                    shape = (int(size),)
                trx.groups[base] = _create_memmap(
                    filename, mode="r+", offset=mem_adress, shape=shape, dtype=ext[1:]
                )
            else:
                logging.error(
                    "{} is not part of a valid structure.".format(elem_filename)
                )

        # All essential array must be declared
        if positions is not None and offsets is not None:
            trx.streamlines._data = positions
            trx.streamlines._offsets = offsets[:-1]
            trx.streamlines._lengths = lengths
        else:
            raise ValueError("Missing essential data.")

        for dpv_key in trx.data_per_vertex:
            tmp = trx.data_per_vertex[dpv_key]
            trx.data_per_vertex[dpv_key] = ArraySequence()
            trx.data_per_vertex[dpv_key]._data = tmp
            trx.data_per_vertex[dpv_key]._offsets = offsets[:-1]
            trx.data_per_vertex[dpv_key]._lengths = lengths
        return trx

    def resize(  # noqa: C901
        self,
        nb_streamlines: Optional[int] = None,
        nb_vertices: Optional[int] = None,
        delete_dpg: bool = False,
    ) -> None:
        """Remove the unused portion of preallocated memmaps.

        Parameters
        ----------
        nb_streamlines : int, optional
            The number of streamlines to keep.
        nb_vertices : int, optional
            The number of vertices to keep.
        delete_dpg : bool, optional
            Remove data_per_group when resizing. Default is False.
        """
        if not self._copy_safe:
            raise ValueError("Cannot resize a sliced datasets.")

        strs_end, pts_end = self._get_real_len()

        if nb_streamlines is not None and nb_streamlines < strs_end:
            strs_end = nb_streamlines
            logging.info(
                "Resizing (down) memmaps, less streamlines than it actually contains."
            )

        if nb_vertices is None:
            pts_end = int(np.sum(self.streamlines._lengths[0:nb_streamlines]))
            nb_vertices = pts_end
        elif nb_vertices < pts_end:
            # Resizing vertices only is too dangerous, not allowed
            logging.warning("Cannot resize (down) vertices for consistency.")
            return

        if nb_streamlines is None:
            nb_streamlines = strs_end

        if (
            nb_streamlines == self.header["NB_STREAMLINES"]
            and nb_vertices == self.header["NB_VERTICES"]
        ):
            logging.debug("TrxFile of the right size, no resizing.")
            return

        trx = self._initialize_empty_trx(nb_streamlines, nb_vertices, init_as=self)

        logging.info(
            "Resizing streamlines from size {} to {}".format(
                len(self.streamlines), nb_streamlines
            )
        )
        logging.info(
            "Resizing vertices from size {} to {}".format(
                len(self.streamlines._data), nb_vertices
            )
        )

        # Copy the fixed-sized info from the original TrxFile to the new
        # (resized) one.
        if nb_streamlines < self.header["NB_STREAMLINES"]:
            trx._copy_fixed_arrays_from(self, nb_strs_to_copy=nb_streamlines)
        else:
            trx._copy_fixed_arrays_from(self)

        tmp_dir = trx._uncompressed_folder_handle.name
        if len(self.groups.keys()) > 0:
            os.mkdir(os.path.join(tmp_dir, "groups/"))

        for group_key in self.groups.keys():
            group_dtype = self.groups[group_key].dtype
            group_name = os.path.join(
                tmp_dir, "groups/", "{}.{}".format(group_key, group_dtype.name)
            )
            ori_len = len(self.groups[group_key])

            # Remove groups indices if resizing down
            tmp = self.groups[group_key][self.groups[group_key] < strs_end]
            trx.groups[group_key] = _create_memmap(
                group_name, mode="w+", shape=(len(tmp),), dtype=group_dtype
            )
            logging.debug(
                "{} group went from {} items to {}".format(group_key, ori_len, len(tmp))
            )
            trx.groups[group_key][:] = tmp

        if delete_dpg:
            self.close()
            self.__dict__ = trx.__dict__
            return

        if len(self.data_per_group.keys()) > 0:
            os.mkdir(os.path.join(tmp_dir, "dpg/"))
        for group_key in self.data_per_group:
            if not os.path.isdir(os.path.join(tmp_dir, "dpg/", group_key)):
                os.mkdir(os.path.join(tmp_dir, "dpg/", group_key))
            if group_key not in trx.data_per_group:
                trx.data_per_group[group_key] = {}

            for dpg_key in self.data_per_group[group_key].keys():
                dpg_dtype = self.data_per_group[group_key][dpg_key].dtype
                dpg_filename = _generate_filename_from_data(
                    self.data_per_group[group_key][dpg_key],
                    os.path.join(tmp_dir, "dpg/", group_key, dpg_key),
                )

                shape = self.data_per_group[group_key][dpg_key].shape
                if dpg_key not in trx.data_per_group[group_key]:
                    trx.data_per_group[group_key][dpg_key] = {}
                trx.data_per_group[group_key][dpg_key] = _create_memmap(
                    dpg_filename, mode="w+", shape=shape, dtype=dpg_dtype
                )

                trx.data_per_group[group_key][dpg_key][:] = self.data_per_group[
                    group_key
                ][dpg_key]

        self.close()
        self.__dict__ = trx.__dict__

    def get_dtype_dict(self):
        """Get the dtype dictionary for the TrxFile.

        Returns
        -------
        dict
            A dictionary containing the dtype for each data element.
        """
        dtype_dict = {
            "positions": self.streamlines._data.dtype,
            "offsets": self.streamlines._offsets.dtype,
            "dpv": {},
            "dps": {},
            "dpg": {},
            "groups": {},
        }

        for key in self.data_per_vertex.keys():
            dtype_dict["dpv"][key] = self.data_per_vertex[key]._data.dtype
        for key in self.data_per_streamline.keys():
            dtype_dict["dps"][key] = self.data_per_streamline[key].dtype

        for group_key in self.data_per_group.keys():
            dtype_dict["groups"][group_key] = self.groups[group_key].dtype

        for group_key in self.data_per_group.keys():
            dtype_dict["dpg"][group_key] = {}
            for dpg_key in self.data_per_group[group_key].keys():
                dtype_dict["dpg"][group_key][dpg_key] = self.data_per_group[group_key][
                    dpg_key
                ].dtype

        return dtype_dict

    def append(self, obj, extra_buffer: int = 0) -> None:
        """Append another tractogram-like object to this TRX.

        Parameters
        ----------
        obj : TrxFile or Tractogram or StatefulTractogram class instance
            Object whose streamlines and associated data will be appended.
        extra_buffer : int, optional
            Additional preallocation buffer for streamlines (in count).

        Returns
        -------
        None
            Mutates the current TrxFile in-place.
        """
        curr_dtype_dict = self.get_dtype_dict()
        if dipy_available:
            from dipy.io.stateful_tractogram import StatefulTractogram

        if not isinstance(obj, (TrxFile, Tractogram)) and (
            dipy_available and not isinstance(obj, StatefulTractogram)
        ):
            raise TypeError(
                "{} is not a supported object type for appending.".format(type(obj))
            )
        elif isinstance(obj, Tractogram):
            obj = self.from_tractogram(
                obj, reference=self.header, dtype_dict=curr_dtype_dict
            )
        elif dipy_available and isinstance(obj, StatefulTractogram):
            obj = self.from_sft(obj, dtype_dict=curr_dtype_dict)

        self._append_trx(obj, extra_buffer=extra_buffer)

    def _append_trx(self, trx: Type["TrxFile"], extra_buffer: int = 0) -> None:
        """Append a TrxFile to another (with buffer support).

        Parameters
        ----------
        trx : TrxFile
            The TrxFile to append to the current TrxFile.
        extra_buffer : int, optional
            The additional buffer space required to append data. Default is 0.
        """
        strs_end, pts_end = self._get_real_len()

        nb_streamlines = strs_end + trx.header["NB_STREAMLINES"]
        nb_vertices = pts_end + trx.header["NB_VERTICES"]

        if (
            self.header["NB_STREAMLINES"] < nb_streamlines
            or self.header["NB_VERTICES"] < nb_vertices
        ):
            self.resize(
                nb_streamlines=nb_streamlines + extra_buffer,
                nb_vertices=nb_vertices + extra_buffer * 100,
            )
        _ = concatenate([self, trx], preallocation=True, delete_groups=True)

    def get_group(
        self, key: str, keep_group: bool = True, copy_safe: bool = False
    ) -> Type["TrxFile"]:
        """Get a particular group from the TrxFile.

        Parameters
        ----------
        key : str
            The group name to select.
        keep_group : bool, optional
            Make sure group exists in returned TrxFile. Default is True.
        copy_safe : bool, optional
            Perform a deepcopy. Default is False.

        Returns
        -------
        TrxFile
            A TrxFile exclusively containing data from said group.
        """
        return self.select(self.groups[key], keep_group=keep_group, copy_safe=copy_safe)

    def select(
        self, indices: np.ndarray, keep_group: bool = True, copy_safe: bool = False
    ) -> Type["TrxFile"]:
        """Get a subset of items, always pointing to the same memmaps.

        Parameters
        ----------
        indices : np.ndarray
            The list of indices of elements to return.
        keep_group : bool, optional
            Ensure group is returned in output TrxFile. Default is True.
        copy_safe : bool, optional
            Perform a deep-copy. Default is False.

        Returns
        -------
        TrxFile
            A TrxFile containing data originating from the selected indices.
        """
        indices = np.array(indices, dtype=np.uint32)

        new_trx = TrxFile()
        new_trx._copy_safe = copy_safe
        new_trx.header = deepcopy(self.header)

        if isinstance(indices, np.ndarray) and len(indices) == 0:
            # Even while empty, basic dtype and header must be coherent
            positions_dtype = self.streamlines._data.dtype
            offsets_dtype = self.streamlines._offsets.dtype
            lengths_dtype = self.streamlines._lengths.dtype
            new_trx.streamlines._data = new_trx.streamlines._data.reshape(
                (0, 3)
            ).astype(positions_dtype)
            new_trx.streamlines._offsets = new_trx.streamlines._offsets.astype(
                offsets_dtype
            )
            new_trx.streamlines._lengths = new_trx.streamlines._lengths.astype(
                lengths_dtype
            )
            new_trx.header["NB_VERTICES"] = len(new_trx.streamlines._data)
            new_trx.header["NB_STREAMLINES"] = len(new_trx.streamlines._lengths)

            return new_trx.deepcopy() if copy_safe else new_trx

        new_trx.streamlines = (
            self.streamlines[indices].copy() if copy_safe else self.streamlines[indices]
        )
        for dpv_key in self.data_per_vertex.keys():
            new_trx.data_per_vertex[dpv_key] = (
                self.data_per_vertex[dpv_key][indices].copy()
                if copy_safe
                else self.data_per_vertex[dpv_key][indices]
            )

        for dps_key in self.data_per_streamline.keys():
            new_trx.data_per_streamline[dps_key] = (
                self.data_per_streamline[dps_key][indices].copy()
                if copy_safe
                else self.data_per_streamline[dps_key][indices]
            )

        # Not keeping group is equivalent to the [] operator
        if keep_group:
            logging.warning("Keeping dpg despite affecting the group items.")
            for group_key in self.groups.keys():
                # Keep the group indices even when fancy slicing
                index = np.argsort(indices)
                sorted_x = indices[index]
                sorted_index = np.searchsorted(sorted_x, self.groups[group_key])
                yindex = np.take(index, sorted_index, mode="clip")
                mask = indices[yindex] != self.groups[group_key]
                intersect = yindex[~mask]

                if len(intersect) == 0:
                    continue

                new_trx.groups[group_key] = intersect
                if group_key in self.data_per_group:
                    for dpg_key in self.data_per_group[group_key].keys():
                        if group_key not in new_trx.data_per_group:
                            new_trx.data_per_group[group_key] = {}
                        new_trx.data_per_group[group_key][dpg_key] = (
                            self.data_per_group[group_key][dpg_key]
                        )

        new_trx.header["NB_VERTICES"] = len(new_trx.streamlines._data)
        new_trx.header["NB_STREAMLINES"] = len(new_trx.streamlines._lengths)
        return new_trx.deepcopy() if copy_safe else new_trx

    @staticmethod
    def from_lazy_tractogram(
        obj: ["LazyTractogram"],
        reference,
        extra_buffer: int = 0,
        chunk_size: int = 10000,
        dtype_dict: dict = None,
    ) -> Type["TrxFile"]:
        """Create a TrxFile from a LazyTractogram with buffer support.

        Parameters
        ----------
        obj : LazyTractogram
            The LazyTractogram to convert.
        reference : object
            Reference for spatial information.
        extra_buffer : int, optional
            The buffer space between reallocation. This number should be a
            number of streamlines. Use 0 for no buffer. Default is 0.
        chunk_size : int, optional
            The number of streamlines to save at a time. Default is 10000.
        dtype_dict : dict, optional
            Dictionary specifying dtypes for positions, offsets, dpv, and dps.

        Returns
        -------
        TrxFile
            A TrxFile created from the LazyTractogram.
        """
        if dtype_dict is None:
            dtype_dict = {
                "positions": np.float32,
                "offsets": np.uint32,
                "dpv": {},
                "dps": {},
            }

        data = {"strs": [], "dpv": {}, "dps": {}}
        concat = None
        count = 0
        iterator = iter(obj)
        while True:
            if count < chunk_size:
                try:
                    i = next(iterator)
                    count += 1
                except StopIteration:
                    obj = convert_data_dict_to_tractogram(data)
                    if concat is None:
                        if len(obj.streamlines) == 0:
                            concat = TrxFile()
                        else:
                            concat = TrxFile.from_tractogram(
                                obj, reference=reference, dtype_dict=dtype_dict
                            )
                    elif len(obj.streamlines) > 0:
                        curr_obj = TrxFile.from_tractogram(
                            obj, reference=reference, dtype_dict=dtype_dict
                        )
                        concat.append(curr_obj)
                    break
                append_generator_to_dict(i, data)
            else:
                obj = convert_data_dict_to_tractogram(data)
                if concat is None:
                    concat = TrxFile.from_tractogram(
                        obj, reference=reference, dtype_dict=dtype_dict
                    )
                else:
                    curr_obj = TrxFile.from_tractogram(
                        obj, reference=reference, dtype_dict=dtype_dict
                    )
                    concat.append(curr_obj, extra_buffer=extra_buffer)
                data = {"strs": [], "dpv": {}, "dps": {}}
                count = 0

        concat.resize()
        return concat

    @staticmethod
    def from_sft(sft, dtype_dict=None):
        """Generate a TrxFile from a StatefulTractogram.

        Parameters
        ----------
        sft : StatefulTractogram class instance
            Input tractogram.
        dtype_dict : dict or None, optional
            Mapping of target dtypes for positions, offsets, dpv, and dps. When
            None, uses ``sft.dtype_dict`` or sensible defaults.

        Returns
        -------
        TrxFile
            TRX representation of the StatefulTractogram.
        """
        if dtype_dict is None:
            dtype_dict = {}

        if len(sft.dtype_dict) > 0:
            dtype_dict = sft.dtype_dict
        if "dpp" in dtype_dict:
            dtype_dict["dpv"] = dtype_dict.pop("dpp")
        elif len(dtype_dict) == 0:
            dtype_dict = {
                "positions": np.float32,
                "offsets": np.uint32,
                "dpv": {},
                "dps": {},
            }

        positions_dtype = dtype_dict["positions"]
        offsets_dtype = dtype_dict["offsets"]

        if not np.issubdtype(positions_dtype, np.floating):
            logging.warning(
                "Casting positions as {}, considering using a floating point "
                "dtype.".format(positions_dtype)
            )

        if not np.issubdtype(offsets_dtype, np.integer):
            logging.warning(
                "Casting offsets as {}, considering using a integer dtype.".format(
                    offsets_dtype
                )
            )

        trx = TrxFile(
            nb_vertices=len(sft.streamlines._data), nb_streamlines=len(sft.streamlines)
        )
        trx.header = {
            "DIMENSIONS": sft.dimensions.tolist(),
            "VOXEL_TO_RASMM": sft.affine.tolist(),
            "NB_VERTICES": len(sft.streamlines._data),
            "NB_STREAMLINES": len(sft.streamlines),
        }

        old_space = deepcopy(sft.space)
        old_origin = deepcopy(sft.origin)

        # TrxFile are written on disk in RASMM/center convention
        sft.to_rasmm()
        sft.to_center()

        tmp_streamlines = deepcopy(sft.streamlines)

        # Cast the int64 of Nibabel to uint32
        tmp_streamlines._offsets = tmp_streamlines._offsets.astype(offsets_dtype)
        tmp_streamlines._data = tmp_streamlines._data.astype(positions_dtype)

        trx.streamlines = tmp_streamlines
        for key in sft.data_per_point:
            dtype_to_use = (
                dtype_dict["dpv"][key] if key in dtype_dict["dpv"] else np.float32
            )
            trx.data_per_vertex[key] = sft.data_per_point[key]
            trx.data_per_vertex[key]._data = sft.data_per_point[key]._data.astype(
                dtype_to_use
            )

        for key in sft.data_per_streamline:
            dtype_to_use = (
                dtype_dict["dps"][key] if key in dtype_dict["dps"] else np.float32
            )
            trx.data_per_streamline[key] = sft.data_per_streamline[key].astype(
                dtype_to_use
            )

        # For safety and for RAM, convert the whole object to memmaps
        tmp_dir = get_trx_tmp_dir()
        save(trx, tmp_dir.name)
        trx.close()
        trx = load_from_directory(tmp_dir.name)
        trx._uncompressed_folder_handle = tmp_dir

        sft.to_space(old_space)
        sft.to_origin(old_origin)
        del tmp_streamlines

        return trx

    @staticmethod
    def from_tractogram(
        tractogram,
        reference,
        dtype_dict=None,
    ):
        """Generate a TrxFile from a nibabel Tractogram.

        Parameters
        ----------
        tractogram : nibabel.streamlines.Tractogram class instance
            Input tractogram to convert.
        reference : object
            Reference anatomy used to populate header fields.
        dtype_dict : dict or None, optional
            Mapping of target dtypes for positions, offsets, dpv, and dps.

        Returns
        -------
        TrxFile class instance
            TRX representation of the tractogram.
        """
        if dtype_dict is None:
            dtype_dict = {
                "positions": np.float32,
                "offsets": np.uint32,
                "dpv": {},
                "dps": {},
            }

        positions_dtype = (
            dtype_dict["positions"] if "positions" in dtype_dict else np.float32
        )
        offsets_dtype = dtype_dict["offsets"] if "offsets" in dtype_dict else np.uint32

        if not np.issubdtype(positions_dtype, np.floating):
            logging.warning(
                "Casting positions as {}, considering using a floating point "
                "dtype.".format(positions_dtype)
            )

        if not np.issubdtype(offsets_dtype, np.integer):
            logging.warning(
                "Casting offsets as {}, considering using a integer dtype.".format(
                    offsets_dtype
                )
            )

        trx = TrxFile(
            nb_vertices=len(tractogram.streamlines._data),
            nb_streamlines=len(tractogram.streamlines),
        )

        affine, dimensions, _, _ = get_reference_info_wrapper(reference)
        trx.header = {
            "DIMENSIONS": dimensions,
            "VOXEL_TO_RASMM": affine,
            "NB_VERTICES": len(tractogram.streamlines._data),
            "NB_STREAMLINES": len(tractogram.streamlines),
        }

        tmp_streamlines = deepcopy(tractogram.streamlines)

        # Cast the int64 of Nibabel to uint32
        tmp_streamlines._offsets = tmp_streamlines._offsets.astype(offsets_dtype)
        tmp_streamlines._data = tmp_streamlines._data.astype(positions_dtype)

        trx.streamlines = tmp_streamlines
        for key in tractogram.data_per_point:
            dtype_to_use = (
                dtype_dict["dpv"][key] if key in dtype_dict["dpv"] else np.float32
            )
            trx.data_per_vertex[key] = tractogram.data_per_point[key]
            trx.data_per_vertex[key]._data = tractogram.data_per_point[
                key
            ]._data.astype(dtype_to_use)

        for key in tractogram.data_per_streamline:
            dtype_to_use = (
                dtype_dict["dps"][key] if key in dtype_dict["dps"] else np.float32
            )
            trx.data_per_streamline[key] = tractogram.data_per_streamline[key].astype(
                dtype_to_use
            )

        # For safety and for RAM, convert the whole object to memmaps
        tmp_dir = get_trx_tmp_dir()
        save(trx, tmp_dir.name)
        trx.close()

        trx = load_from_directory(tmp_dir.name)
        del tmp_streamlines

        return trx

    def to_tractogram(self, resize=False):
        """Convert this TrxFile to a nibabel Tractogram.

        Parameters
        ----------
        resize : bool, optional
            If True, resize to actual data length before conversion.

        Returns
        -------
        nibabel.streamlines.Tractogram class instance
            Tractogram containing streamlines and metadata.
        """
        if resize:
            self.resize()

        trx_obj = self.to_memory()
        tractogram = nib.streamlines.Tractogram([], affine_to_rasmm=np.eye(4))
        tractogram._set_streamlines(trx_obj.streamlines)
        tractogram._data_per_point = trx_obj.data_per_vertex
        tractogram._data_per_streamline = trx_obj.data_per_streamline

        return tractogram

    def to_memory(self, resize: bool = False) -> Type["TrxFile"]:
        """Convert a TrxFile to a RAM representation.

        Parameters
        ----------
        resize : bool, optional
            Resize TrxFile when converting to RAM representation.
            Default is False.

        Returns
        -------
        TrxFile
            A non memory-mapped TrxFile.
        """
        if resize:
            self.resize()

        trx_obj = TrxFile()
        trx_obj.header = deepcopy(self.header)
        trx_obj.streamlines = deepcopy(self.streamlines)

        for key in self.data_per_vertex:
            trx_obj.data_per_vertex[key] = deepcopy(self.data_per_vertex[key])

        for key in self.data_per_streamline:
            trx_obj.data_per_streamline[key] = deepcopy(self.data_per_streamline[key])

        for key in self.groups:
            trx_obj.groups[key] = deepcopy(self.groups[key])

        for key in self.data_per_group:
            trx_obj.data_per_group[key] = deepcopy(self.data_per_group[key])

        return trx_obj

    def to_sft(self, resize=False):
        """Convert this TrxFile to a StatefulTractogram.

        Parameters
        ----------
        resize : bool, optional
            If True, resize to actual data length before conversion.

        Returns
        -------
        StatefulTractogram class instance or None
            StatefulTractogram object, or None if dipy is unavailable.
        """
        try:
            from dipy.io.stateful_tractogram import Space, StatefulTractogram
        except ImportError:
            logging.error(
                "Dipy library is missing, cannot convert to StatefulTractogram."
            )
            return None

        affine = np.array(self.header["VOXEL_TO_RASMM"], dtype=np.float32)
        dimensions = np.array(self.header["DIMENSIONS"], dtype=np.uint16)
        vox_sizes = np.array(voxel_sizes(affine), dtype=np.float32)
        vox_order = "".join(aff2axcodes(affine))
        space_attributes = (affine, dimensions, vox_sizes, vox_order)

        if resize:
            self.resize()
        sft = StatefulTractogram(
            deepcopy(self.streamlines),
            space_attributes,
            Space.RASMM,
            data_per_point=deepcopy(self.data_per_vertex),
            data_per_streamline=deepcopy(self.data_per_streamline),
        )
        tmp_dict = self.get_dtype_dict()
        if "dpv" in tmp_dict:
            tmp_dict["dpp"] = tmp_dict.pop("dpv")
        sft.dtype_dict = self.get_dtype_dict()

        return sft

    def close(self) -> None:
        """Cleanup on-disk temporary folder and memmaps.

        Returns
        -------
        None
            Releases file handles and removes temporary storage.
        """
        close_or_delete_mmap(self.streamlines)

        for key in self.data_per_vertex:
            close_or_delete_mmap(self.data_per_vertex[key])

        for key in self.data_per_streamline:
            close_or_delete_mmap(self.data_per_streamline[key])

        for key in self.groups:
            close_or_delete_mmap(self.groups[key])

        for key in self.data_per_group:
            for dpg in self.data_per_group[key]:
                close_or_delete_mmap(self.data_per_group[key][dpg])

        if self._uncompressed_folder_handle is not None:
            try:
                self._uncompressed_folder_handle.cleanup()
            except PermissionError:
                logging.error(
                    "Windows PermissionError, temporary directory %s was not deleted!",
                    self._uncompressed_folder_handle.name,
                )
        self.__init__()
        logging.debug("Deleted memmaps and initialized empty TrxFile.")



"""Sphinx configuration for trx-python documentation."""
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import os
from datetime import datetime as dt

# -- Version information -----------------------------------------------------
# Get version from environment variable (set by CI) or package
version = os.environ.get('TRX_VERSION', None)
if version is None:
    try:
        from trx import __version__
        version = __version__
    except ImportError:
        version = "dev"

# Normalize version for switcher matching
# Remove .devX suffix for matching against switcher.json
version_match = version.split('.dev')[0] if '.dev' in version else version
if version_match == version and 'dev' not in version:
    # This is a release version
    pass
else:
    # Development version - match against "dev"
    version_match = "dev"

# -- Project information -----------------------------------------------------

project = 'trx-python'
copyright = copyright = f'2021-{dt.now().year}, The TRX developers'
author = 'The TRX developers'
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.autosummary',
    'autoapi.extension',
    'numpydoc',
    'sphinx_gallery.gen_gallery',
    'sphinx_design',
]

# Suppress known deprecation warnings from dependencies
# astroid 4.x deprecation - will be fixed when sphinx-autoapi updates for astroid 5.x
import warnings
warnings.filterwarnings(
    'ignore',
    message="importing .* from 'astroid' is deprecated",
    category=DeprecationWarning
)

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['../_static']
html_logo = "../_static/trx_logo.png"

html_sidebars = {
    "scripts": [],
    "trx_specifications": [],
}

html_theme_options = {
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/tee-ar-ex",  # required
            # Icon class (if "type": "fontawesome"), or path to local image
            # (if "type": "local")
            "icon": "fab fa-github-square",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        }
    ],
    # Version switcher configuration
    "switcher": {
        "json_url": "https://tee-ar-ex.github.io/trx-python/switcher.json",
        "version_match": version_match,
    },
    "navbar_start": ["navbar-logo", "version-switcher"],
    "show_version_warning_banner": True,
    # Show table of contents on each page (section navigation)
    "secondary_sidebar_items": ["page-toc", "edit-this-page", "sourcelink"],
    "show_toc_level": 2,
}


autoapi_type = 'python'
autoapi_dirs = ['../../trx']
autoapi_ignore = ['*test*', '*version*']


def _validate_reference_urls(urls, timeout=5):
    """Validate reference URLs and return only reachable ones.

    Checks if the objects.inv file (used by sphinx for intersphinx) is
    accessible at each URL.

    Parameters
    ----------
    urls : dict
        Dictionary of package names to documentation URLs.
    timeout : int
        Connection timeout in seconds.

    Returns
    -------
    dict
        Dictionary containing only URLs that are reachable.
    """
    import urllib.request
    import urllib.error

    valid_urls = {}
    for name, url in urls.items():
        objects_inv_url = url.rstrip('/') + '/objects.inv'
        try:
            req = urllib.request.Request(
                objects_inv_url,
                headers={'User-Agent': 'Sphinx-doc-builder'}
            )
            urllib.request.urlopen(req, timeout=timeout)
            valid_urls[name] = url
        except urllib.error.URLError as e:
            reason = getattr(e, 'reason', str(e))
            print(f"WARNING: Skipping '{name}' reference URL ({url}): {reason}")
        except Exception as e:
            print(f"WARNING: Skipping '{name}' reference URL ({url}): {e}")
    return valid_urls


# Reference URLs for sphinx-gallery hyperlinks
_reference_urls = {
    'numpy': 'https://numpy.org/doc/stable/',
    'nibabel': 'https://nipy.org/nibabel/',
}

# Sphinx gallery configuration
sphinx_gallery_conf = {
     'examples_dirs': '../../examples',
     'gallery_dirs': 'auto_examples',
     'within_subsection_order': 'NumberOfCodeLinesSortKey',
     'reference_url': _validate_reference_urls(_reference_urls),
     'default_thumb_file': os.path.join(os.path.dirname(__file__), '..', '_static', 'trx_logo.png'),
}



# -*- coding: utf-8 -*-
# sphinx_gallery_thumbnail_path = '../docs/_static/trx_logo.png'
"""
Data Per Vertex and Data Per Streamline
========================================

This tutorial demonstrates how to work with metadata in TRX files.
TRX supports two types of metadata:

- **Data Per Vertex (dpv)**: Information attached to each point along streamlines
- **Data Per Streamline (dps)**: Information attached to entire streamlines

By the end of this tutorial, you will know how to:

- Access dpv and dps data in a TRX file
- Understand the data shapes and organization
- Use metadata for filtering and analysis
"""

# %%
# Understanding DPV and DPS
# -------------------------
#
# **Data Per Vertex (dpv):**
#
# - Attached to each individual point (vertex) in all streamlines
# - Shape: (NB_VERTICES, 1) for scalar data or (NB_VERTICES, N) for vector data
# - Common uses: FA values at each point, RGB colors, local orientations
#
# **Data Per Streamline (dps):**
#
# - Attached to entire streamlines (one value per streamline)
# - Shape: (NB_STREAMLINES, 1) for scalar data or (NB_STREAMLINES, N) for vector data
# - Common uses: bundle ID, mean FA, streamline length, tracking algorithm ID

# %%
# Loading a TRX file with metadata
# --------------------------------
#
# Let's load a TRX file and explore its metadata.

import os

import numpy as np

from trx.fetcher import fetch_data, get_home, get_testing_files_dict
from trx.trx_file_memmap import load

# Download test data
fetch_data(get_testing_files_dict(), keys="gold_standard.zip")
trx_home = get_home()
trx_path = os.path.join(trx_home, "gold_standard", "gs.trx")

# Load the TRX file
trx = load(trx_path)

print(f"Loaded TRX with {len(trx)} streamlines")
print(f"Total vertices: {trx.header['NB_VERTICES']}")

# %%
# Exploring Data Per Vertex (dpv)
# -------------------------------
#
# Let's see what dpv data is available.

print("Data Per Vertex keys:", list(trx.data_per_vertex.keys()))

# Examine each dpv field
for key in trx.data_per_vertex:
    data = trx.data_per_vertex[key]
    print(f"\n  {key}:")
    print(f"    Shape: {data._data.shape}")
    print(f"    Dtype: {data._data.dtype}")
    print(f"    Sample values: {data._data[:3].flatten()}")

# %%
# Accessing dpv for a specific streamline
# ---------------------------------------
#
# The dpv data is organized to match the streamlines. You can access
# the dpv values for a specific streamline using the same indices.

if len(trx.data_per_vertex) > 0:
    first_dpv_key = list(trx.data_per_vertex.keys())[0]
    dpv_data = trx.data_per_vertex[first_dpv_key]

    # Get dpv values for the first streamline
    first_streamline_dpv = dpv_data[0]
    print(f"DPV '{first_dpv_key}' for first streamline:")
    print(f"  Shape: {first_streamline_dpv.shape}")
    print(f"  Values: {first_streamline_dpv.flatten()}")

# %%
# Exploring Data Per Streamline (dps)
# -----------------------------------
#
# Now let's examine the dps data.

print("Data Per Streamline keys:", list(trx.data_per_streamline.keys()))

# Examine each dps field
for key in trx.data_per_streamline:
    data = trx.data_per_streamline[key]
    print(f"\n  {key}:")
    print(f"    Shape: {data.shape}")
    print(f"    Dtype: {data.dtype}")
    print(f"    First 5 values: {data[:5].flatten()}")

# %%
# DPS for filtering streamlines
# -----------------------------
#
# A common use case is filtering streamlines based on dps values.
# For example, selecting streamlines with high FA values.

if len(trx.data_per_streamline) > 0:
    # Use the first dps key for demonstration
    first_dps_key = list(trx.data_per_streamline.keys())[0]
    dps_data = trx.data_per_streamline[first_dps_key]

    # Calculate some statistics
    print(f"\nStatistics for '{first_dps_key}':")
    print(f"  Min: {np.min(dps_data):.4f}")
    print(f"  Max: {np.max(dps_data):.4f}")
    print(f"  Mean: {np.mean(dps_data):.4f}")
    print(f"  Std: {np.std(dps_data):.4f}")

# %%
# File structure for dpv and dps
# ------------------------------
#
# In the TRX format, dpv and dps are stored in separate directories:
#
# .. code-block:: text
#
#     my_tractogram.trx/
#     |-- dpv/
#     |   |-- fa.float16              # FA values per vertex
#     |   |-- colors.3.uint8          # RGB colors (3 values per vertex)
#     |   +-- curvature.float32       # Curvature per vertex
#     |-- dps/
#     |   |-- bundle_id.uint8         # Bundle assignment per streamline
#     |   |-- length.uint16           # Length per streamline
#     |   +-- mean_fa.float32         # Mean FA per streamline
#     +-- ...
#
# The filename format is: ``name.dtype`` or ``name.dimension.dtype``

# %%
# Working with multi-dimensional data
# -----------------------------------
#
# Both dpv and dps can have multiple dimensions. For example, RGB colors
# have 3 values per vertex.

print("\nDemonstrating multi-dimensional data:")

# Check for any multi-dimensional dpv
for key in trx.data_per_vertex:
    data = trx.data_per_vertex[key]
    if len(data._data.shape) > 1 and data._data.shape[1] > 1:
        print(f"  {key}: {data._data.shape[1]}D data per vertex")

# Check for any multi-dimensional dps
for key in trx.data_per_streamline:
    data = trx.data_per_streamline[key]
    if len(data.shape) > 1 and data.shape[1] > 1:
        print(f"  {key}: {data.shape[1]}D data per streamline")

# %%
# Relationship between dpv and streamlines
# ----------------------------------------
#
# It's important to understand how dpv data maps to individual streamlines.
# Each streamline's dpv values can be accessed using the streamline's
# vertex indices.

# Get vertex counts for first few streamlines
print("\nVertex distribution for first 5 streamlines:")
for i in range(min(5, len(trx))):
    streamline = trx.streamlines[i]
    print(f"  Streamline {i}: {len(streamline)} vertices")

# Total vertices should match
total_from_streamlines = sum(len(trx.streamlines[i]) for i in range(len(trx)))
print(f"\nTotal vertices from streamlines: {total_from_streamlines}")
print(f"Total vertices in header: {trx.header['NB_VERTICES']}")

# %%
# Summary
# -------
#
# In this tutorial, you learned how to:
#
# - Access dpv data using ``trx.data_per_vertex[key]``
# - Access dps data using ``trx.data_per_streamline[key]``
# - Understand the shape conventions for scalar and vector data
# - Use metadata for statistical analysis
# - Understand the file structure for dpv and dps
#
# The TRX format's metadata system is designed for flexibility, allowing
# you to attach any kind of information to vertices or streamlines.



# -*- coding: utf-8 -*-
# sphinx_gallery_thumbnail_path = '../docs/_static/trx_logo.png'
"""
Working with Groups
====================

This tutorial demonstrates how to work with groups in TRX files.
Groups allow you to organize streamlines into meaningful subsets,
such as anatomical bundles or clusters.

By the end of this tutorial, you will know how to:

- Access groups in a TRX file
- Extract streamlines belonging to a specific group
- Understand the relationship between groups and data_per_group (dpg)
- Work with overlapping groups
"""

# %%
# What are Groups?
# ----------------
#
# Groups in TRX files are collections of streamline indices. They enable:
#
# - **Sparse representation**: Only store indices instead of copying data
# - **Overlapping membership**: A streamline can belong to multiple groups
# - **Efficient access**: Quickly extract predefined subsets of streamlines
#
# Common use cases include anatomical bundles (e.g., Arcuate Fasciculus,
# Corpus Callosum), clustering results, or connectivity-based groupings.

# %%
# Loading a TRX file with groups
# ------------------------------
#
# Let's load a TRX file that contains group information.

import os

import numpy as np

from trx.fetcher import fetch_data, get_home, get_testing_files_dict
from trx.trx_file_memmap import load

# Download test data
fetch_data(get_testing_files_dict(), keys="gold_standard.zip")
trx_home = get_home()
trx_path = os.path.join(trx_home, "gold_standard", "gs.trx")

# Load the TRX file
trx = load(trx_path)

print(f"Loaded TRX with {len(trx)} streamlines")

# %%
# Accessing groups
# ----------------
#
# Groups are stored as a dictionary where keys are group names and values
# are numpy arrays of streamline indices.

print(f"Available groups: {list(trx.groups.keys())}")

# Check the number of groups
print(f"Number of groups: {len(trx.groups)}")

# %%
# Let's examine the groups in more detail:

for group_name, indices in trx.groups.items():
    print(f"  {group_name}: {len(indices)} streamlines")

# %%
# Extracting a group
# ------------------
#
# You can extract all streamlines belonging to a specific group using
# the ``get_group()`` method.

if len(trx.groups) > 0:
    # Get the first group name
    first_group = list(trx.groups.keys())[0]

    # Extract the group as a new TrxFile
    group_trx = trx.get_group(first_group)
    print(f"Extracted group '{first_group}' with {len(group_trx)} streamlines")

    # You can also access the raw indices
    group_indices = trx.groups[first_group]
    print(f"Raw indices (first 10): {group_indices[:10]}")
else:
    print("No groups available in this file")

# %%
# Using group indices directly
# ----------------------------
#
# You can use group indices to select streamlines directly with the
# ``select()`` method.

if len(trx.groups) > 0:
    first_group = list(trx.groups.keys())[0]
    indices = trx.groups[first_group]

    # Select streamlines using indices
    selected = trx.select(indices[:5])  # Select first 5 from the group
    print(f"Selected {len(selected)} streamlines from group '{first_group}'")

# %%
# Data per group (dpg)
# --------------------
#
# Groups can have associated metadata stored in ``data_per_group`` (dpg).
# This is useful for storing group-level statistics like mean FA, volume,
# or color codes.

print(f"Data per group keys: {list(trx.data_per_group.keys())}")

# Check what metadata is available for each group
for group_name in trx.data_per_group:
    dpg_keys = list(trx.data_per_group[group_name].keys())
    print(f"  {group_name}: {dpg_keys}")

# %%
# Creating groups manually
# ------------------------
#
# You can create groups by assigning indices to the groups dictionary.
# Here's an example of how groups work conceptually.

# Example: Create conceptual groups for 10 streamlines
example_groups = {
    "bundle_A": np.array([0, 1, 2, 3], dtype=np.uint32),
    "bundle_B": np.array([4, 5, 6, 7, 8, 9], dtype=np.uint32),
    "overlapping": np.array([3, 4, 5], dtype=np.uint32),  # Overlaps with A and B
}

print("Example groups:")
for name, indices in example_groups.items():
    print(f"  {name}: streamlines {indices}")

# Note: Streamline 3 is in both bundle_A and overlapping
# Note: Streamlines 4, 5 are in both bundle_B and overlapping
print("\nOverlapping groups are allowed in TRX!")

# %%
# Group file structure
# --------------------
#
# In the TRX file format, groups are stored as binary files in a ``groups/``
# directory:
#
# .. code-block:: text
#
#     my_tractogram.trx/
#     |-- groups/
#     |   |-- AF_L.uint32      # Arcuate Fasciculus Left
#     |   |-- AF_R.uint32      # Arcuate Fasciculus Right
#     |   |-- CC.uint32        # Corpus Callosum
#     |   +-- CST_L.uint32     # Corticospinal Tract Left
#     +-- ...
#
# Each file contains a flat array of streamline indices as uint32 values.

# %%
# Filtering streamlines by group
# ------------------------------
#
# A common workflow is to filter streamlines based on group membership
# and then analyze or visualize specific bundles.

if len(trx.groups) > 0:
    # Get all group names
    group_names = list(trx.groups.keys())

    # Report statistics for each group
    print("Group statistics:")
    for group_name in group_names:
        group_trx = trx.get_group(group_name)
        total_points = len(group_trx.streamlines._data)
        avg_length = total_points / len(group_trx) if len(group_trx) > 0 else 0
        print(f"  {group_name}:")
        print(f"    - Streamlines: {len(group_trx)}")
        print(f"    - Total points: {total_points}")
        print(f"    - Avg points per streamline: {avg_length:.1f}")

# %%
# Summary
# -------
#
# In this tutorial, you learned how to:
#
# - Access groups using ``trx.groups``
# - Extract group streamlines using ``get_group()``
# - Work with ``data_per_group`` (dpg) metadata
# - Understand that groups can overlap
# - Filter and analyze streamlines by group membership
#
# Groups are a powerful feature of the TRX format that enable efficient
# organization and retrieval of streamline subsets without data duplication.



"""Custom spin commands for trx-python development."""

import glob
import os
import shutil
import subprocess
import sys
import tempfile

import click

UPSTREAM_URL = "https://github.com/tee-ar-ex/trx-python.git"
UPSTREAM_NAME = "upstream"


def run(cmd, check=True, capture=True):
    """Run a shell command.

    Parameters
    ----------
    cmd : list of str
        Command and arguments to execute.
    check : bool, optional
        If True, check the return code and report errors.
    capture : bool, optional
        If True, capture stdout and stderr.

    Returns
    -------
    str or int or None
        Captured stdout string, return code, or None on error.
    """
    result = subprocess.run(cmd, capture_output=capture, text=True, check=False)
    if check and result.returncode != 0:
        if capture:
            click.echo(f"Error: {result.stderr}", err=True)
        return None
    return result.stdout.strip() if capture else result.returncode


def get_remotes():
    """Get dict of remote names to URLs.

    Returns
    -------
    dict
        Mapping of remote names to their fetch URLs.
    """
    output = run(["git", "remote", "-v"])
    if not output:
        return {}
    remotes = {}
    for line in output.split("\n"):
        if "(fetch)" in line:
            parts = line.split()
            remotes[parts[0]] = parts[1]
    return remotes


@click.command()
def setup():
    """Set up development environment (fetch tags from upstream).

    This command configures your fork for development by:
    1. Adding the upstream remote if not present
    2. Fetching tags from upstream (required for correct version detection)

    Run this once after cloning your fork.
    """
    click.echo("Setting up trx-python development environment...\n")

    # Check if in git repo
    if run(["git", "rev-parse", "--git-dir"], check=False) is None:
        click.echo("Error: Not in a git repository", err=True)
        sys.exit(1)

    # Check/add upstream remote
    remotes = get_remotes()
    upstream_remote = None

    for name, url in remotes.items():
        if UPSTREAM_URL.rstrip(".git") in url.rstrip(".git"):
            upstream_remote = name
            click.echo(f"Found upstream remote: {name}")
            break

    if upstream_remote is None:
        click.echo(f"Adding upstream remote: {UPSTREAM_URL}")
        run(["git", "remote", "add", UPSTREAM_NAME, UPSTREAM_URL])
        upstream_remote = UPSTREAM_NAME

    # Fetch tags
    click.echo(f"\nFetching tags from {upstream_remote}...")
    run(["git", "fetch", upstream_remote, "--tags"], capture=False)

    # Verify version
    click.echo("\nVerifying version detection...")
    try:
        from setuptools_scm import get_version

        version = get_version()
        click.echo(f"Detected version: {version}")

        # Check for suspicious version patterns
        if version.startswith("0.0"):
            click.echo(
                "\nWarning: Version starts with 0.0 - tags may not be fetched.",
                err=True,
            )
            sys.exit(1)
    except ImportError:
        click.echo("Note: Install setuptools_scm to verify version detection")

    click.echo("\nSetup complete! You can now run:")
    click.echo("  spin install    # Install in development mode")
    click.echo("  spin test       # Run tests")


@click.command()
@click.option(
    "-m",
    "--match",
    "pattern",
    default=None,
    help="Only run tests matching this pattern (passed to pytest -k)",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbose output")
@click.argument("pytest_args", nargs=-1)
def test(pattern, verbose, pytest_args):
    """Run tests using pytest.

    Additional arguments are passed directly to pytest.

    Parameters
    ----------
    pattern : str or None
        Only run tests matching this pattern (passed to pytest -k).
    verbose : bool
        If True, enable verbose output.
    pytest_args : tuple
        Additional arguments passed directly to pytest.
    """
    cmd = ["pytest", "trx/tests"]

    if pattern:
        cmd.extend(["-k", pattern])

    if verbose:
        cmd.append("-v")

    if pytest_args:
        cmd.extend(pytest_args)

    click.echo(f"Running: {' '.join(cmd)}\n")
    sys.exit(run(cmd, capture=False, check=False))


@click.command()
@click.option(
    "--fix", is_flag=True, default=False, help="Automatically fix issues where possible"
)
def lint(fix):
    """Run linting checks using ruff and codespell.

    Parameters
    ----------
    fix : bool
        If True, automatically fix issues where possible.
    """
    click.echo("Running ruff linter...")
    cmd = ["ruff", "check", "."]

    if fix:
        cmd.append("--fix")

    result = run(cmd, capture=False, check=False)
    if result != 0:
        click.echo("\nLinting issues found!", err=True)
        sys.exit(1)

    click.echo("\nRunning ruff formatter check...")
    cmd_format = ["ruff", "format", "--check", "."]
    result = run(cmd_format, capture=False, check=False)
    if result != 0:
        click.echo("\nFormatting issues found!", err=True)
        sys.exit(1)

    click.echo("\nRunning codespell...")
    cmd_spell = [
        "codespell",
        "--skip",
        "*.pyc,.git,pyproject.toml,./docs/_build/*,*.egg-info,./build/*,./dist/*,./tmp/*",
        "trx",
        "docs/source",
        ".spin",
    ]
    result = run(cmd_spell, capture=False, check=False)
    if result != 0:
        click.echo("\nSpelling issues found!", err=True)
        sys.exit(1)

    click.echo("\nAll checks passed!")


@click.command()
@click.option(
    "--clean", is_flag=True, default=False, help="Clean build directory before building"
)
@click.option(
    "--open",
    "open_browser",
    is_flag=True,
    default=False,
    help="Open documentation in browser after building",
)
def docs(clean, open_browser):
    """Build documentation using Sphinx.

    Parameters
    ----------
    clean : bool
        If True, clean build directory before building.
    open_browser : bool
        If True, open documentation in browser after building.
    """
    import os

    docs_dir = "docs"

    if clean:
        click.echo("Cleaning build directory...")
        build_dir = os.path.join(docs_dir, "_build")
        if os.path.exists(build_dir):
            shutil.rmtree(build_dir)

        # Clean sphinx-gallery generated files
        gallery_dir = os.path.join(docs_dir, "source", "auto_examples")
        if os.path.exists(gallery_dir):
            click.echo("Cleaning sphinx-gallery generated files...")
            shutil.rmtree(gallery_dir)

        # Clean sphinx-gallery execution times file
        sg_times = os.path.join(docs_dir, "source", "sg_execution_times.rst")
        if os.path.exists(sg_times):
            os.remove(sg_times)

    click.echo("Building documentation...")
    cmd = ["make", "-C", docs_dir, "html"]
    result = run(cmd, capture=False, check=False)

    if result == 0:
        index_path = os.path.abspath(
            os.path.join(docs_dir, "_build", "html", "index.html")
        )
        click.echo("\nDocs built successfully!")
        click.echo(f"Open: {index_path}")

        if open_browser:
            import webbrowser

            webbrowser.open(f"file://{index_path}")

    sys.exit(result)


@click.command()
def clean():  # noqa: C901
    """Clean up temporary files and build artifacts."""
    click.echo("Cleaning up temporary files...")

    # Clean TRX temp directory
    trx_tmp_dir = os.getenv("TRX_TMPDIR", tempfile.gettempdir())
    if os.path.exists(trx_tmp_dir):
        temp_files = glob.glob(os.path.join(trx_tmp_dir, "trx_*"))
        for temp_dir in temp_files:
            if os.path.isdir(temp_dir):
                click.echo(f"Removing temporary directory: {temp_dir}")
                shutil.rmtree(temp_dir)

    # Clean build artifacts
    for build_pattern in ["build", "dist", "*.egg-info"]:
        for path in glob.glob(build_pattern):
            if os.path.isdir(path):
                click.echo(f"Removing build directory: {path}")
                shutil.rmtree(path)
            elif os.path.isfile(path):
                click.echo(f"Removing build file: {path}")
                os.remove(path)

    # Clean Python cache
    for cache_dir in ["**/__pycache__", "**/.pytest_cache"]:
        for path in glob.glob(cache_dir, recursive=True):
            if os.path.isdir(path):
                click.echo(f"Removing cache directory: {path}")
                shutil.rmtree(path)

    click.echo("Cleanup complete!")



# -*- coding: utf-8 -*-
# sphinx_gallery_thumbnail_path = '../docs/_static/trx_logo.png'
"""
Reading and Writing TRX Files
==============================

This tutorial demonstrates how to read and write TRX files using trx-python.
TRX is a tractography file format designed for efficient storage and access
of brain fiber tract streamline data.

By the end of this tutorial, you will know how to:

- Load a TRX file from disk
- Inspect the contents of a TRX file
- Access streamlines and metadata
- Save a TRX file to disk
- Create a TRX file from scratch
"""

# %%
# Loading a TRX file
# ------------------
#
# Let's start by loading an existing TRX file. First, we need to download
# some test data.

import os
import tempfile

from trx.fetcher import fetch_data, get_home, get_testing_files_dict
from trx.trx_file_memmap import load, save

# Download test data
fetch_data(get_testing_files_dict(), keys="gold_standard.zip")
trx_home = get_home()
trx_path = os.path.join(trx_home, "gold_standard", "gs.trx")

# Load the TRX file
trx = load(trx_path)

print("TRX file loaded successfully!")

# %%
# Inspecting TRX file contents
# ----------------------------
#
# The TrxFile object has several key attributes that you can inspect.
# Let's look at what's inside our loaded file.

# Print a summary of the TRX file
print(trx)

# %%
# The header contains essential metadata about the tractogram:

print("Header information:")
print(f"  Number of streamlines: {trx.header['NB_STREAMLINES']}")
print(f"  Number of vertices: {trx.header['NB_VERTICES']}")
print(f"  Image dimensions: {trx.header['DIMENSIONS']}")
print(f"  Voxel to RASMM affine:\n{trx.header['VOXEL_TO_RASMM']}")

# %%
# Accessing streamlines
# ---------------------
#
# Streamlines are the core data in a TRX file. Each streamline is a sequence
# of 3D points representing a fiber tract in the brain.

print(f"Number of streamlines: {len(trx)}")
print(f"Total number of vertices: {len(trx.streamlines._data)}")

# Access the first streamline
first_streamline = trx.streamlines[0]
print(f"\nFirst streamline has {len(first_streamline)} points")
print(f"First 3 points of the first streamline:\n{first_streamline[:3]}")

# %%
# Accessing metadata
# ------------------
#
# TRX files can contain additional data per vertex (dpv) and per streamline (dps).

print("Data per vertex (dpv) keys:", list(trx.data_per_vertex.keys()))
print("Data per streamline (dps) keys:", list(trx.data_per_streamline.keys()))
print("Groups:", list(trx.groups.keys()))

# %%
# Selecting a subset of streamlines
# ---------------------------------
#
# You can easily select a subset of streamlines using indices or slicing.

# Select first 5 streamlines
subset = trx[:5]
print(f"Subset has {len(subset)} streamlines")

# Select specific streamlines by indices (ensure indices are valid)
max_idx = len(trx) - 1
indices = [0, min(2, max_idx), min(5, max_idx)]
selected = trx.select(indices)
print(f"Selected {len(selected)} streamlines")

# %%
# Saving a TRX file
# -----------------
#
# You can save a TRX file back to disk. The file can be saved as a compressed
# or uncompressed zip archive, or as a directory.

with tempfile.TemporaryDirectory() as tmpdir:
    # Save as TRX file (zip archive)
    output_path = os.path.join(tmpdir, "output.trx")
    save(trx, output_path)
    print(f"Saved TRX file to: {output_path}")
    print(f"File size: {os.path.getsize(output_path)} bytes")

    # Reload to verify
    reloaded = load(output_path)
    print(f"Reloaded TRX has {len(reloaded)} streamlines")

# %%
# Creating a TRX file from an existing one
# ----------------------------------------
#
# A common workflow is to create a new TRX file based on an existing one,
# preserving the spatial reference information.

# Create a deepcopy of the loaded TRX file
trx_copy = trx.deepcopy()

print(f"Created copy with {len(trx_copy)} streamlines")
print(f"Header preserved: DIMENSIONS = {trx_copy.header['DIMENSIONS']}")

# %%
# Summary
# -------
#
# In this tutorial, you learned how to:
#
# - Load TRX files using ``load()``
# - Inspect header information and streamline data
# - Access data per vertex (dpv) and data per streamline (dps)
# - Select subsets of streamlines
# - Save TRX files using ``save()``
# - Create copies of TRX files using ``deepcopy()``
#
# The TRX format is designed for memory efficiency through memory-mapping,
# making it suitable for large tractography datasets.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Update switcher.json for documentation version switching.

This script maintains the version switcher JSON file used by pydata-sphinx-theme
to enable users to switch between different documentation versions.
"""

import argparse
import json
from pathlib import Path
import sys

BASE_URL = "https://tee-ar-ex.github.io/trx-python"


def load_switcher(path):
    """Load existing switcher.json or return empty list.

    Parameters
    ----------
    path : str or Path
        Path to the switcher.json file.

    Returns
    -------
    list
        List of version entries from the switcher file.
    """
    try:
        with open(path, "r") as f:
            return json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        return []


def save_switcher(path, versions):
    """Save switcher.json with proper formatting.

    Parameters
    ----------
    path : str or Path
        Path to the switcher.json file.
    versions : list
        List of version entries to write.
    """
    with open(path, "w") as f:
        json.dump(versions, f, indent=4)
        f.write("\n")


def ensure_dev_entry(versions):
    """Ensure dev entry exists in versions list.

    Parameters
    ----------
    versions : list
        List of version entries.

    Returns
    -------
    list
        Updated versions list with dev entry.
    """
    dev_exists = any(v.get("version") == "dev" for v in versions)
    if not dev_exists:
        versions.insert(0, {"name": "dev", "version": "dev", "url": f"{BASE_URL}/dev/"})
    return versions


def ensure_stable_entry(versions):
    """Ensure stable entry exists with preferred flag.

    Parameters
    ----------
    versions : list
        List of version entries.

    Returns
    -------
    list
        Updated versions list with stable entry.
    """
    stable_idx = next(
        (i for i, v in enumerate(versions) if v.get("version") == "stable"), None
    )
    if stable_idx is not None:
        versions[stable_idx]["preferred"] = True
    else:
        versions.append(
            {
                "name": "stable",
                "version": "stable",
                "url": f"{BASE_URL}/stable/",
                "preferred": True,
            }
        )
    return versions


def add_version(versions, version):
    """Add a new version entry to the versions list.

    Parameters
    ----------
    versions : list
        List of version entries.
    version : str
        Version string to add (e.g., "0.5.0").

    Returns
    -------
    list
        Updated versions list.
    """
    # Remove 'preferred' from all existing entries
    for v in versions:
        v.pop("preferred", None)

    # Check if this version already exists
    version_exists = any(v.get("version") == version for v in versions)

    if not version_exists:
        new_entry = {
            "name": version,
            "version": version,
            "url": f"{BASE_URL}/{version}/",
        }
        # Find dev entry index to insert after it
        dev_idx = next(
            (i for i, v in enumerate(versions) if v.get("version") == "dev"), -1
        )
        if dev_idx >= 0:
            versions.insert(dev_idx + 1, new_entry)
        else:
            versions.insert(0, new_entry)

    return versions


def main():
    """Run the switcher update workflow.

    Returns
    -------
    int
        Exit code (0 for success).
    """
    parser = argparse.ArgumentParser(
        description="Update switcher.json for documentation version switching"
    )
    parser.add_argument("switcher_path", type=Path, help="Path to switcher.json file")
    parser.add_argument("--version", type=str, help="New version to add (e.g., 0.5.0)")

    args = parser.parse_args()

    # Load existing versions
    versions = load_switcher(args.switcher_path)

    # Add new version if specified
    if args.version:
        versions = add_version(versions, args.version)

    # Ensure required entries exist
    versions = ensure_dev_entry(versions)
    versions = ensure_stable_entry(versions)

    # Save updated switcher.json
    save_switcher(args.switcher_path, versions)

    # Print result for CI logs
    print(f"Updated {args.switcher_path}:")
    print(json.dumps(versions, indent=4))

    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
"""High-level processing workflows for tractogram operations."""

from copy import deepcopy
import csv
import gzip
import json
import logging
import os
import tempfile

import nibabel as nib
from nibabel.streamlines.array_sequence import ArraySequence
import numpy as np

try:
    import dipy  # noqa: F401

    dipy_available = True
except ImportError:
    dipy_available = False

from trx.io import get_trx_tmp_dir, load, load_sft_with_reference, save
from trx.streamlines_ops import intersection, perform_streamlines_operation
import trx.trx_file_memmap as tmm
from trx.utils import (
    flip_sft,
    get_axis_shift_vector,
    get_reference_info_wrapper,
    get_reverse_enum,
    is_header_compatible,
    load_matrix_in_any_format,
    split_name_with_gz,
)
from trx.viz import display


def convert_dsi_studio(
    in_dsi_tractogram,
    in_dsi_fa,
    out_tractogram,
    remove_invalid=True,
    keep_invalid=False,
):
    """Convert a DSI-Studio TRK file to TRX, fixing space metadata.

    Parameters
    ----------
    in_dsi_tractogram : str
        Input DSI-Studio TRK path (optionally .trk.gz).
    in_dsi_fa : str
        FA image (.nii.gz) used as reference anatomy.
    out_tractogram : str
        Destination tractogram path; ``.trx`` will be written using TRX writer.
    remove_invalid : bool, optional
        Remove streamlines falling outside the bounding box. Defaults to True.
    keep_invalid : bool, optional
        Keep invalid streamlines even if outside bounding box. Defaults to False.

    Returns
    -------
    None
        Writes the converted tractogram to disk.
    """
    if not dipy_available:
        logging.error("Dipy library is missing, scripts are not available.")
        return None
    from dipy.io.stateful_tractogram import Space, StatefulTractogram
    from dipy.io.streamline import load_tractogram, save_tractogram

    in_ext = split_name_with_gz(in_dsi_tractogram)[1]
    out_ext = split_name_with_gz(out_tractogram)[1]

    if in_ext == ".trk.gz":
        with gzip.open(in_dsi_tractogram, "rb") as f_in:
            with open("tmp.trk", "wb") as f_out:
                f_out.writelines(f_in)
                sft = load_tractogram("tmp.trk", "same", bbox_valid_check=False)
            os.remove("tmp.trk")
    elif in_ext == ".trk":
        sft = load_tractogram(in_dsi_tractogram, "same", bbox_valid_check=False)
    else:
        raise IOError("{} is not currently supported.".format(in_ext))

    sft.to_vox()
    sft_fix = StatefulTractogram(
        sft.streamlines,
        in_dsi_fa,
        Space.VOXMM,
        data_per_point=sft.data_per_point,
        data_per_streamline=sft.data_per_streamline,
    )
    sft_fix.to_vox()
    flip_axis = ["x", "y"]
    sft_fix.streamlines._data -= get_axis_shift_vector(flip_axis)
    sft_flip = flip_sft(sft_fix, flip_axis)

    sft_flip.to_rasmm()
    sft_flip.streamlines._data -= [0.5, 0.5, -0.5]

    if remove_invalid:
        sft_flip.remove_invalid_streamlines()

    if out_ext != ".trx":
        save_tractogram(sft_flip, out_tractogram, bbox_valid_check=not keep_invalid)
    else:
        trx = tmm.TrxFile.from_sft(sft_flip)
        tmm.save(trx, out_tractogram)


def convert_tractogram(  # noqa: C901
    in_tractogram,
    out_tractogram,
    reference,
    pos_dtype="float32",
    offsets_dtype="uint32",
):
    """Convert tractograms between formats with dtype control.

    Parameters
    ----------
    in_tractogram : str
        Input tractogram path.
    out_tractogram : str
        Output tractogram path.
    reference : str
        Reference anatomy required for formats without header affine.
    pos_dtype : str, optional
        Datatype for positions in TRX output.
    offsets_dtype : str, optional
        Datatype for offsets in TRX output.

    Returns
    -------
    None
        Writes the converted tractogram to disk.
    """
    if not dipy_available:
        logging.error("Dipy library is missing, scripts are not available.")
        return None
    from dipy.io.streamline import save_tractogram

    in_ext = split_name_with_gz(in_tractogram)[1]
    out_ext = split_name_with_gz(out_tractogram)[1]

    if in_ext == out_ext:
        raise IOError("Input and output cannot be of the same file format.")

    if in_ext != ".trx":
        sft = load_sft_with_reference(in_tractogram, reference, bbox_check=False)
    else:
        trx = tmm.load(in_tractogram)
        sft = trx.to_sft()
        trx.close()

    if out_ext != ".trx":
        if out_ext == ".vtk":
            if sft.streamlines._data.dtype.name != pos_dtype:
                sft.streamlines._data = sft.streamlines._data.astype(pos_dtype)
            if offsets_dtype == "uint64" or offsets_dtype == "uint32":
                offsets_dtype = offsets_dtype[1:]
            if sft.streamlines._offsets.dtype.name != offsets_dtype:
                sft.streamlines._offsets = sft.streamlines._offsets.astype(
                    offsets_dtype
                )
        save_tractogram(sft, out_tractogram, bbox_valid_check=False)
    else:
        trx = tmm.TrxFile.from_sft(sft)
        if trx.streamlines._data.dtype.name != pos_dtype:
            trx.streamlines._data = trx.streamlines._data.astype(pos_dtype)
        if trx.streamlines._offsets.dtype.name != offsets_dtype:
            trx.streamlines._offsets = trx.streamlines._offsets.astype(offsets_dtype)
        tmm.save(trx, out_tractogram)
        trx.close()


def tractogram_simple_compare(in_tractograms, reference):
    """Compare tractograms against a reference and return a summary diff.

    Parameters
    ----------
    in_tractograms : list of str
        Paths to tractograms to compare.
    reference : str
        Reference tractogram path.

    Returns
    -------
    dict
        Dictionary capturing differences across tractograms.
    """
    if not dipy_available:
        logging.error("Dipy library is missing, scripts are not available.")
        return
    from dipy.io.stateful_tractogram import StatefulTractogram

    tractogram_obj = load(in_tractograms[0], reference)
    if not isinstance(tractogram_obj, StatefulTractogram):
        sft_1 = tractogram_obj.to_sft()
        tractogram_obj.close()
    else:
        sft_1 = tractogram_obj

    tractogram_obj = load(in_tractograms[1], reference)
    if not isinstance(tractogram_obj, StatefulTractogram):
        sft_2 = tractogram_obj.to_sft()
        tractogram_obj.close()
    else:
        sft_2 = tractogram_obj

    if np.allclose(sft_1.streamlines._data, sft_2.streamlines._data, atol=0.001):
        print("Matching tractograms in rasmm!")
    else:
        print(
            "Average difference in rasmm of {}".format(
                np.average(sft_1.streamlines._data - sft_2.streamlines._data, axis=0)
            )
        )

    sft_1.to_voxmm()
    sft_2.to_voxmm()
    if np.allclose(sft_1.streamlines._data, sft_2.streamlines._data, atol=0.001):
        print("Matching tractograms in voxmm!")
    else:
        print(
            "Average difference in voxmm of {}".format(
                np.average(sft_1.streamlines._data - sft_2.streamlines._data, axis=0)
            )
        )

    sft_1.to_vox()
    sft_2.to_vox()
    if np.allclose(sft_1.streamlines._data, sft_2.streamlines._data, atol=0.001):
        print("Matching tractograms in vox!")
    else:
        print(
            "Average difference in vox of {}".format(
                np.average(sft_1.streamlines._data - sft_2.streamlines._data, axis=0)
            )
        )


def verify_header_compatibility(in_files):
    """Verify that multiple tractogram headers are mutually compatible.

    Parameters
    ----------
    in_files : list of str
        Paths to tractogram or NIfTI files to compare.

    Returns
    -------
    None
        Prints compatibility results to stdout.
    """
    if not dipy_available:
        logging.error("Dipy library is missing, scripts are not available.")
        return

    all_valid = True
    for filepath in in_files:
        if not os.path.isfile(filepath):
            print("{} does not exist".format(filepath))
        _, in_extension = split_name_with_gz(filepath)
        if in_extension not in [".trk", ".nii", ".nii.gz", ".trx"]:
            raise IOError("{} does not have a supported extension".format(filepath))
        if not is_header_compatible(in_files[0], filepath):
            print(
                "{} and {} do not have compatible header.".format(in_files[0], filepath)
            )
            all_valid = False
    if all_valid:
        print("All input files have compatible headers.")


def tractogram_visualize_overlap(in_tractogram, reference, remove_invalid=True):
    """Visualize overlap between tractogram density maps in different spaces.

    Parameters
    ----------
    in_tractogram : str
        Input tractogram path.
    reference : str
        Reference anatomy (.nii or .nii.gz).
    remove_invalid : bool, optional
        Remove streamlines outside bounding box before visualization.

    Returns
    -------
    None
        Opens interactive windows when fury is available.
    """
    if not dipy_available:
        logging.error("Dipy library is missing, scripts are not available.")
        return None
    from dipy.io.stateful_tractogram import StatefulTractogram
    from dipy.tracking.streamline import set_number_of_points
    from dipy.tracking.utils import density_map

    tractogram_obj = load(in_tractogram, reference)
    if not isinstance(tractogram_obj, StatefulTractogram):
        sft = tractogram_obj.to_sft()
        tractogram_obj.close()
    else:
        sft = tractogram_obj
    sft.streamlines._data = sft.streamlines._data.astype(float)

    sft.data_per_point = None
    sft.streamlines = set_number_of_points(sft.streamlines, 200)

    if remove_invalid:
        sft.remove_invalid_streamlines()

    # Approach (1)
    density_1 = density_map(sft.streamlines, sft.affine, sft.dimensions)
    img = nib.load(reference)
    display(
        img.get_fdata(),
        volume_affine=img.affine,
        streamlines=sft.streamlines,
        title="RASMM",
    )

    # Approach (2)
    sft.to_vox()
    density_2 = density_map(sft.streamlines, np.eye(4), sft.dimensions)

    # Small difference due to casting of the affine as float32 or float64
    diff = density_1 - density_2
    print(
        "Total difference of {} voxels with total value of {}".format(
            np.count_nonzero(diff), np.sum(np.abs(diff))
        )
    )

    display(img.get_fdata(), streamlines=sft.streamlines, title="VOX")

    # Try VOXMM
    sft.to_voxmm()
    affine = np.eye(4)
    affine[0:3, 0:3] *= sft.voxel_sizes

    display(
        img.get_fdata(),
        volume_affine=affine,
        streamlines=sft.streamlines,
        title="VOXMM",
    )


def validate_tractogram(
    in_tractogram,
    reference,
    out_tractogram,
    remove_identical_streamlines=True,
    precision=1,
):
    """Validate a tractogram and optionally remove invalid/duplicate streamlines.

    Parameters
    ----------
    in_tractogram : str
        Input tractogram path.
    reference : str
        Reference anatomy for formats requiring it.
    out_tractogram : str or None
        Optional output path to save the cleaned tractogram.
    remove_identical_streamlines : bool, optional
        Remove duplicate streamlines based on hashing precision.
    precision : int, optional
        Number of decimals when hashing streamline points.

    Returns
    -------
    None
        Prints warnings and optionally writes a cleaned tractogram.
    """
    if not dipy_available:
        logging.error("Dipy library is missing, scripts are not available.")
        return None
    from dipy.io.stateful_tractogram import StatefulTractogram

    tractogram_obj = load(in_tractogram, reference)

    if not isinstance(tractogram_obj, StatefulTractogram):
        sft = tractogram_obj.to_sft()
        # tractogram_obj.close()
    else:
        sft = tractogram_obj

    ori_dtype = sft.dtype_dict
    ori_len = len(sft)
    tot_remove = 0

    invalid_coord_ind, _ = sft.remove_invalid_streamlines()
    tot_remove += len(invalid_coord_ind)
    logging.warning(
        "Removed {} streamlines with invalid coordinates.".format(
            len(invalid_coord_ind)
        )
    )

    indices = [i for i in range(len(sft)) if len(sft.streamlines[i]) <= 1]
    tot_remove = +len(indices)
    logging.warning(
        "Removed {} invalid streamlines (1 or 0 points).".format(len(indices))
    )

    for i in np.setdiff1d(range(len(sft)), indices):
        norm = np.linalg.norm(np.diff(sft.streamlines[i], axis=0), axis=1)

        if (norm < 0.001).any():
            indices.append(i)

    indices_val = np.setdiff1d(range(len(sft)), indices).astype(np.uint32)
    logging.warning(
        "Removed {} invalid streamlines (overlapping points).".format(
            ori_len - len(indices_val)
        )
    )
    tot_remove += ori_len - len(indices_val)

    if remove_identical_streamlines:
        _, indices_uniq = perform_streamlines_operation(
            intersection, [sft.streamlines], precision=precision
        )
        indices_final = np.intersect1d(indices_val, indices_uniq).astype(np.uint32)
        logging.warning(
            "Removed {} overlapping streamlines.".format(
                ori_len - len(indices_final) - tot_remove
            )
        )

        indices_final = np.intersect1d(indices_val, indices_uniq)
    else:
        indices_final = indices_val

    if out_tractogram:
        streamlines = sft.streamlines[indices_final].copy()
        dpp = {}
        for key in sft.data_per_point.keys():
            dpp[key] = sft.data_per_point[key][indices_final].copy()

        dps = {}
        for key in sft.data_per_streamline.keys():
            dps[key] = sft.data_per_streamline[key][indices_final]
        new_sft = StatefulTractogram.from_sft(
            streamlines, sft, data_per_point=dpp, data_per_streamline=dps
        )
        new_sft.dtype_dict = ori_dtype
        save(new_sft, out_tractogram)


def _load_streamlines_from_csv(positions_csv):
    """Load streamlines from a CSV file.

    Parameters
    ----------
    positions_csv : str
        Path to CSV containing flattened coordinates.

    Returns
    -------
    nibabel.streamlines.ArraySequence class instance
        Streamlines reconstructed from the CSV rows.
    """
    with open(positions_csv, newline="") as f:
        reader = csv.reader(f)
        data = list(reader)
        data = [np.reshape(i, (len(i) // 3, 3)).astype(float) for i in data]
        return ArraySequence(data)


def _load_streamlines_from_arrays(positions, offsets):
    """Load streamlines from position and offset arrays.

    Parameters
    ----------
    positions : str
        Path to positions array (.npy or text) shaped (N, 3).
    offsets : str
        Path to offsets array marking streamline boundaries.

    Returns
    -------
    tuple
        (ArraySequence, np.ndarray) of streamlines and offsets.
    """
    positions = load_matrix_in_any_format(positions)
    offsets = load_matrix_in_any_format(offsets)
    lengths = tmm._compute_lengths(offsets)
    streamlines = ArraySequence()
    streamlines._data = positions
    streamlines._offsets = deepcopy(offsets)
    streamlines._lengths = lengths
    return streamlines, offsets


def _apply_spatial_transforms(
    streamlines, reference, space_str, origin_str, verify_invalid, offsets
):
    """Apply spatial transforms and optionally remove invalid streamlines.

    Parameters
    ----------
    streamlines : ArraySequence class instance
        Streamlines to transform.
    reference : str
        Reference anatomy used for space/origin.
    space_str : str
        Desired space (e.g., \"rasmm\").
    origin_str : str
        Desired origin (e.g., \"nifti\").
    verify_invalid : bool
        Remove streamlines outside bounding box when True.
    offsets : np.ndarray
        Offsets array to preserve after transforms.

    Returns
    -------
    ArraySequence class instance or None
        Transformed streamlines, or None if dipy is unavailable.
    """
    if not dipy_available:
        logging.error(
            "Dipy library is missing, advanced options "
            "related to spatial transforms and invalid "
            "streamlines are not available."
        )
        return None

    from dipy.io.stateful_tractogram import StatefulTractogram

    space, origin = get_reverse_enum(space_str, origin_str)
    sft = StatefulTractogram(streamlines, reference, space, origin)
    if verify_invalid:
        rem, _ = sft.remove_invalid_streamlines()
        print(
            "{} streamlines were removed becaused they were invalid.".format(len(rem))
        )
    sft.to_rasmm()
    sft.to_center()
    streamlines = sft.streamlines
    streamlines._offsets = offsets
    return streamlines


def _write_header(tmp_dir_name, reference, streamlines):
    """Write TRX header file to a temporary directory.

    Parameters
    ----------
    tmp_dir_name : str
        Temporary directory where header.json is written.
    reference : str
        Reference anatomy used to derive affine and dimensions.
    streamlines : ArraySequence class instance
        Streamlines whose counts populate the header.
    """
    affine, dimensions, _, _ = get_reference_info_wrapper(reference)
    header = {
        "DIMENSIONS": dimensions.tolist(),
        "VOXEL_TO_RASMM": affine.tolist(),
        "NB_VERTICES": len(streamlines._data),
        "NB_STREAMLINES": len(streamlines) - 1,
    }

    if header["NB_STREAMLINES"] <= 1:
        raise IOError("To use this script, you need at least 2streamlines.")

    with open(os.path.join(tmp_dir_name, "header.json"), "w") as out_json:
        json.dump(header, out_json)


def _write_streamline_data(tmp_dir_name, streamlines, positions_dtype, offsets_dtype):
    """Write streamline position and offset data.

    Parameters
    ----------
    tmp_dir_name : str
        Temporary directory to store binary arrays.
    streamlines : ArraySequence class instance
        Streamlines to serialize.
    positions_dtype : str
        Datatype for positions array.
    offsets_dtype : str
        Datatype for offsets array.
    """
    curr_filename = os.path.join(tmp_dir_name, "positions.3.{}".format(positions_dtype))
    positions = streamlines._data.astype(positions_dtype)
    tmm._ensure_little_endian(positions).tofile(curr_filename)

    curr_filename = os.path.join(tmp_dir_name, "offsets.{}".format(offsets_dtype))
    offsets = streamlines._offsets.astype(offsets_dtype)
    tmm._ensure_little_endian(offsets).tofile(curr_filename)


def _normalize_dtype(dtype_str):
    """Normalize dtype string format for file naming.

    Parameters
    ----------
    dtype_str : str
        Input dtype string (e.g., \"bool\", \"float32\").

    Returns
    -------
    str
        Normalized dtype string where ``bool`` is mapped to ``bit``.
    """
    return "bit" if dtype_str == "bool" else dtype_str


def _write_data_array(tmp_dir_name, subdir_name, args, is_dpg=False):
    """Write a data array (dpv/dps/group/dpg) to disk.

    Parameters
    ----------
    tmp_dir_name : str
        Base temporary directory.
    subdir_name : str
        Subdirectory name (dpv, dps, groups, dpg).
    args : tuple
        Tuple describing the array path and dtype (and group when dpg).
    is_dpg : bool, optional
        True when writing data_per_group arrays.

    Returns
    -------
    None
        Writes the array to disk.
    """
    if is_dpg:
        os.makedirs(os.path.join(tmp_dir_name, "dpg", args[0]), exist_ok=True)
        curr_arr = load_matrix_in_any_format(args[1]).astype(args[2])
        basename = os.path.basename(os.path.splitext(args[1])[0])
        dtype_str = _normalize_dtype(args[1]) if args[1] != "bool" else "bit"
        dtype = args[2]
    else:
        os.makedirs(os.path.join(tmp_dir_name, subdir_name), exist_ok=True)
        curr_arr = np.squeeze(load_matrix_in_any_format(args[0]).astype(args[1]))
        basename = os.path.basename(os.path.splitext(args[0])[0])
        dtype_str = _normalize_dtype(args[1])
        dtype = dtype_str

    if curr_arr.ndim > 2:
        raise IOError("Maximum of 2 dimensions for dpv/dps/dpg.")

    if curr_arr.shape == (1, 1):
        curr_arr = curr_arr.reshape((1,))

    dim = "" if curr_arr.ndim == 1 else "{}.".format(curr_arr.shape[-1])

    if is_dpg:
        curr_filename = os.path.join(
            tmp_dir_name, "dpg", args[0], "{}.{}{}".format(basename, dim, dtype)
        )
    else:
        curr_filename = os.path.join(
            tmp_dir_name, subdir_name, "{}.{}{}".format(basename, dim, dtype)
        )

    tmm._ensure_little_endian(curr_arr).tofile(curr_filename)


def generate_trx_from_scratch(  # noqa: C901
    reference,
    out_tractogram,
    positions_csv=False,
    positions=False,
    offsets=False,
    positions_dtype="float32",
    offsets_dtype="uint64",
    space_str="rasmm",
    origin_str="nifti",
    verify_invalid=True,
    dpv=None,
    dps=None,
    groups=None,
    dpg=None,
):
    """Generate TRX file from scratch using various input formats.

    Parameters
    ----------
    reference : str
        Reference anatomy used to set affine and dimensions.
    out_tractogram : str
        Output TRX filename.
    positions_csv : str or bool, optional
        CSV file containing streamline coordinates; False to disable.
    positions : str or bool, optional
        Binary positions array file; False to disable.
    offsets : str or bool, optional
        Offsets array file; False to disable.
    positions_dtype : str, optional
        Datatype for positions.
    offsets_dtype : str, optional
        Datatype for offsets.
    space_str : str, optional
        Desired space for generated streamlines.
    origin_str : str, optional
        Desired origin for generated streamlines.
    verify_invalid : bool, optional
        Remove invalid streamlines when True.
    dpv : list or None, optional
        Data per vertex definitions.
    dps : list or None, optional
        Data per streamline definitions.
    groups : list or None, optional
        Group definitions.
    dpg : list or None, optional
        Data per group definitions.

    Returns
    -------
    None
        Writes the generated TRX file to disk.
    """
    if dpv is None:
        dpv = []
    if dps is None:
        dps = []
    if groups is None:
        groups = []
    if dpg is None:
        dpg = []

    with get_trx_tmp_dir() as tmp_dir_name:
        if positions_csv:
            streamlines = _load_streamlines_from_csv(positions_csv)
            offsets = None
        else:
            streamlines, offsets = _load_streamlines_from_arrays(positions, offsets)

        if (
            space_str.lower() != "rasmm"
            or origin_str.lower() != "nifti"
            or verify_invalid
        ):
            streamlines = _apply_spatial_transforms(
                streamlines, reference, space_str, origin_str, verify_invalid, offsets
            )
            if streamlines is None:
                return

        _write_header(tmp_dir_name, reference, streamlines)
        _write_streamline_data(
            tmp_dir_name, streamlines, positions_dtype, offsets_dtype
        )

        if dpv:
            for arg in dpv:
                _write_data_array(tmp_dir_name, "dpv", arg)

        if dps:
            for arg in dps:
                _write_data_array(tmp_dir_name, "dps", arg)

        if groups:
            for arg in groups:
                _write_data_array(tmp_dir_name, "groups", arg)

        if dpg:
            for arg in dpg:
                _write_data_array(tmp_dir_name, "dpg", arg, is_dpg=True)

        trx = tmm.load(tmp_dir_name)
        tmm.save(trx, out_tractogram)
        trx.close()


def manipulate_trx_datatype(in_filename, out_filename, dict_dtype):  # noqa: C901
    """Change dtype of positions, offsets, dpv, dps, dpg, and groups in a TRX.

    Parameters
    ----------
    in_filename : str
        Input TRX file path.
    out_filename : str
        Output TRX file path.
    dict_dtype : dict
        Mapping describing target dtypes for each data category.

    Returns
    -------
    None
        Writes the converted TRX to ``out_filename``.
    """
    trx = tmm.load(in_filename)

    # For each key in dict_dtype, we create a new memmap with the new dtype
    # and we copy the data from the old memmap to the new one.
    for key in dict_dtype:
        if key == "positions":
            tmp_mm = np.memmap(
                tempfile.NamedTemporaryFile(),
                dtype=dict_dtype[key],
                mode="w+",
                shape=trx.streamlines._data.shape,
            )
            tmp_mm[:] = trx.streamlines._data[:]
            trx.streamlines._data = tmp_mm
        elif key == "offsets":
            tmp_mm = np.memmap(
                tempfile.NamedTemporaryFile(),
                dtype=dict_dtype[key],
                mode="w+",
                shape=trx.streamlines._offsets.shape,
            )
            tmp_mm[:] = trx.streamlines._offsets[:]
            trx.streamlines._offsets = tmp_mm
        elif key == "dpv":
            for key_dpv in dict_dtype[key]:
                tmp_mm = np.memmap(
                    tempfile.NamedTemporaryFile(),
                    dtype=dict_dtype[key][key_dpv],
                    mode="w+",
                    shape=trx.data_per_vertex[key_dpv]._data.shape,
                )
                tmp_mm[:] = trx.data_per_vertex[key_dpv]._data[:]
                trx.data_per_vertex[key_dpv]._data = tmp_mm
        elif key == "dps":
            for key_dps in dict_dtype[key]:
                tmp_mm = np.memmap(
                    tempfile.NamedTemporaryFile(),
                    dtype=dict_dtype[key][key_dps],
                    mode="w+",
                    shape=trx.data_per_streamline[key_dps].shape,
                )
                tmp_mm[:] = trx.data_per_streamline[key_dps][:]
                trx.data_per_streamline[key_dps] = tmp_mm
        elif key == "dpg":
            for key_group in dict_dtype[key]:
                for key_dpg in dict_dtype[key][key_group]:
                    tmp_mm = np.memmap(
                        tempfile.NamedTemporaryFile(),
                        dtype=dict_dtype[key][key_group][key_dpg],
                        mode="w+",
                        shape=trx.data_per_group[key_group][key_dpg].shape,
                    )
                    tmp_mm[:] = trx.data_per_group[key_group][key_dpg][:]
                    trx.data_per_group[key_group][key_dpg] = tmp_mm
        elif key == "groups":
            for key_group in dict_dtype[key]:
                tmp_mm = np.memmap(
                    tempfile.NamedTemporaryFile(),
                    dtype=dict_dtype[key][key_group],
                    mode="w+",
                    shape=trx.groups[key_group].shape,
                )
                tmp_mm[:] = trx.groups[key_group][:]
                trx.groups[key_group] = tmp_mm

    tmm.save(trx, out_filename)
    trx.close()



# -*- coding: utf-8 -*-
"""Optional 3D visualization using FURY/VTK."""

import itertools
import logging

import numpy as np

try:
    from dipy.viz import actor, colormap, window
    import fury.utils as ut_vtk
    from fury.utils import get_bounds
    import vtk

    fury_available = True
except ImportError:
    fury_available = False


def display(
    volume, volume_affine=None, streamlines=None, title="FURY", display_bounds=True
):
    """Display a volume with optional streamlines using fury.

    Parameters
    ----------
    volume : np.ndarray
        3D volume to display.
    volume_affine : np.ndarray or None, optional
        Affine matrix for the volume; None assumes identity.
    streamlines : sequence or None, optional
        Streamlines to render as lines.
    title : str, optional
        Window title.
    display_bounds : bool, optional
        If True, draw bounding box and coordinate annotations.

    Returns
    -------
    None
        Opens an interactive visualization window when fury is available.
    """
    if not fury_available:
        logging.error(
            "Fury library is missing, visualization functions are not available."
        )
        return None
    volume = volume.astype(float)
    scene = window.Scene()
    scene.background((1.0, 0.5, 0.0))

    # Show the X/Y/Z plane intersecting, mid-slices
    slicer_actor_1 = actor.slicer(
        volume,
        affine=volume_affine,
        value_range=(volume.min(), volume.max()),
        interpolation="nearest",
        opacity=0.8,
    )
    slicer_actor_2 = actor.slicer(
        volume,
        affine=volume_affine,
        value_range=(volume.min(), volume.max()),
        interpolation="nearest",
        opacity=0.8,
    )
    slicer_actor_3 = actor.slicer(
        volume,
        affine=volume_affine,
        value_range=(volume.min(), volume.max()),
        interpolation="nearest",
        opacity=0.8,
    )
    slicer_actor_1.display(y=volume.shape[1] // 2)
    slicer_actor_2.display(x=volume.shape[0] // 2)
    slicer_actor_3.display(z=volume.shape[2] // 2)

    scene.add(slicer_actor_1)
    scene.add(slicer_actor_2)
    scene.add(slicer_actor_3)

    # Bounding box to facilitate error detections
    if display_bounds:
        src = vtk.vtkCubeSource()
        bounds = np.round(get_bounds(slicer_actor_1), 6)
        src.SetBounds(bounds)
        src.Update()
        cube_actor = ut_vtk.get_actor_from_polydata(src.GetOutput())
        cube_actor.GetProperty().SetRepresentationToWireframe()
        scene.add(cube_actor)

        # Show each corner's coordinates
        corners = itertools.product(bounds[0:2], bounds[2:4], bounds[4:6])
        for corner in corners:
            text_actor = actor.text_3d(
                "{}, {}, {}".format(*corner),
                corner,
                font_size=6,
                justification="center",
            )
            scene.add(text_actor)

        # Show the X/Y/Z dimensions
        text_actor_x = actor.text_3d(
            "{}".format(np.abs(bounds[0] - bounds[1])),
            ((bounds[0] + bounds[1]) / 2, bounds[2], bounds[4]),
            font_size=10,
            justification="center",
        )
        text_actor_y = actor.text_3d(
            "{}".format(np.abs(bounds[2] - bounds[3])),
            (bounds[0], (bounds[2] + bounds[3]) / 2, bounds[4]),
            font_size=10,
            justification="center",
        )
        text_actor_z = actor.text_3d(
            "{}".format(np.abs(bounds[4] - bounds[5])),
            (bounds[0], bounds[2], (bounds[4] + bounds[5]) / 2),
            font_size=10,
            justification="center",
        )
        scene.add(text_actor_x)
        scene.add(text_actor_y)
        scene.add(text_actor_z)

    if streamlines is not None:
        streamlines_actor = actor.line(
            streamlines, colormap.line_colors(streamlines), opacity=0.25
        )
        scene.add(streamlines_actor)
    window.show(scene, title=title, size=(800, 800))



# -*- coding: utf-8 -*-
"""Unified I/O interface for tractogram file formats."""

import logging
import os
import sys
import tempfile

try:
    import dipy  # noqa: F401

    dipy_available = True
except ImportError:
    dipy_available = False

from trx.utils import split_name_with_gz


def get_trx_tmp_dir():
    """Return a temporary directory honoring the ``TRX_TMPDIR`` setting.

    When the ``TRX_TMPDIR`` environment variable is set to ``"use_working_dir"``
    the current working directory is used. Otherwise, the value of
    ``TRX_TMPDIR`` is used directly. If the variable is not set, the system
    temporary directory is used.

    Returns
    -------
    tempfile.TemporaryDirectory
        Context-managed temporary directory placed according to the environment
        configuration.
    """
    if os.getenv("TRX_TMPDIR") is not None:
        if os.getenv("TRX_TMPDIR") == "use_working_dir":
            trx_tmp_dir = os.getcwd()
        else:
            trx_tmp_dir = os.getenv("TRX_TMPDIR")
    else:
        trx_tmp_dir = tempfile.gettempdir()

    if sys.version_info[1] >= 10:
        return tempfile.TemporaryDirectory(
            dir=trx_tmp_dir, prefix="trx_", ignore_cleanup_errors=True
        )
    else:
        return tempfile.TemporaryDirectory(dir=trx_tmp_dir, prefix="trx_")


def load_sft_with_reference(filepath, reference=None, bbox_check=True):
    """Load a tractogram as a StatefulTractogram with an explicit reference.

    Parameters
    ----------
    filepath : str
        Path to the tractogram file (.trk, .tck, .fib, .vtk, .dpy).
    reference : str or nibabel.Nifti1Image, optional
        Reference image used for formats without embedded affine information.
        Pass ``"same"`` to reuse the header embedded in .trk files.
    bbox_check : bool, optional
        If True, validate that streamlines lie within the reference bounding
        box. Defaults to True.

    Returns
    -------
    StatefulTractogram or None
        Loaded tractogram. Returns ``None`` when ``dipy`` is unavailable.

    Raises
    ------
    IOError
        If the file format is unsupported or a required reference is missing.
    """
    if not dipy_available:
        logging.error(
            "Dipy library is missing, cannot use functions related "
            "to the StatefulTractogram."
        )
        return None
    from dipy.io.streamline import load_tractogram

    # Force the usage of --reference for all file formats without an header
    _, ext = os.path.splitext(filepath)
    if ext == ".trk":
        if reference is not None and reference != "same":
            logging.warning(
                "Reference is discarded for this file format {}.".format(filepath)
            )
        sft = load_tractogram(filepath, "same", bbox_valid_check=bbox_check)
    elif ext in [".tck", ".fib", ".vtk", ".dpy"]:
        if reference is None or reference == "same":
            raise IOError(
                "--reference is required for this file format {}.".format(filepath)
            )
        else:
            sft = load_tractogram(filepath, reference, bbox_valid_check=bbox_check)

    else:
        raise IOError("{} is an unsupported file format".format(filepath))

    return sft


def load(tractogram_filename, reference):
    """Load a tractogram from disk and return a TRX or StatefulTractogram.

    Parameters
    ----------
    tractogram_filename : str
        Path to the input tractogram. TRX directories are supported.
    reference : str or nibabel.Nifti1Image
        Reference image used for formats without embedded affine information.

    Returns
    -------
    TrxFile or StatefulTractogram
        TRX file handle for ``.trx`` inputs, otherwise a StatefulTractogram.
    """
    import trx.trx_file_memmap as tmm

    in_ext = split_name_with_gz(tractogram_filename)[1]
    if in_ext != ".trx" and not os.path.isdir(tractogram_filename):
        tractogram_obj = load_sft_with_reference(
            tractogram_filename, reference, bbox_check=False
        )
    else:
        tractogram_obj = tmm.load(tractogram_filename)

    return tractogram_obj


def save(tractogram_obj, tractogram_filename, bbox_valid_check=False):
    """Save a tractogram object to disk.

    Parameters
    ----------
    tractogram_obj : TrxFile or StatefulTractogram
        Tractogram to persist. Non-TRX inputs are converted to StatefulTractogram
        before saving to non-TRX formats.
    tractogram_filename : str
        Destination file name. ``.trx`` will be saved using the TRX writer; all
        other extensions are handled by ``dipy.save_tractogram``.
    bbox_valid_check : bool, optional
        If True, validate that streamlines lie within the reference bounding
        box when saving non-TRX formats. Defaults to False.

    Returns
    -------
    None
        The function writes to disk and returns ``None``. Returns ``None``
        immediately when ``dipy`` is unavailable.
    """
    if not dipy_available:
        logging.error(
            "Dipy library is missing, cannot use functions related "
            "to the StatefulTractogram."
        )
        return None
    from dipy.io.stateful_tractogram import StatefulTractogram
    from dipy.io.streamline import save_tractogram

    import trx.trx_file_memmap as tmm

    out_ext = split_name_with_gz(tractogram_filename)[1]

    if out_ext != ".trx":
        if not isinstance(tractogram_obj, StatefulTractogram):
            tractogram_obj = tractogram_obj.to_sft()
        save_tractogram(
            tractogram_obj, tractogram_filename, bbox_valid_check=bbox_valid_check
        )
    else:
        if not isinstance(tractogram_obj, tmm.TrxFile):
            tractogram_obj = tmm.TrxFile.from_sft(tractogram_obj)
        tmm.save(tractogram_obj, tractogram_filename)
        tractogram_obj.close()
